# -*- coding: utf-8 -*-

"""
<p> ACES Data Processor Module:
</p>
This module provides functions for data input/output and preprocessing for the ACES project.
"""

import logging
logging.basicConfig(level=logging.INFO)

import numpy as np
import tensorflow as tf
from functools import partial

__all__ = ["DataProcessor", "RandomTransform"]

class DataProcessor:
    """
    ACES Data processor Class:

    This class provides functions for data input/output and preprocessing for the ACES project.
    """

    @staticmethod
    @tf.autograph.experimental.do_not_convert
    def create_tfrecord_from_file(filename: str) -> tf.data.TFRecordDataset:
        """
        Create a TensorFlow Dataset from a TFRecord file.

        Parameters:

        * filename (str): The filename of the TFRecord file.

        Returns:

        * tf.data.TFRecordDataset: The TensorFlow Dataset created from the TFRecord file.
        """
        return tf.data.TFRecordDataset(filename, compression_type="GZIP")

    @staticmethod
    @tf.autograph.experimental.do_not_convert
    def get_sum_tensor(records):
        """
        Gets the total number of tensor record by mapping through them.

        Parameters:

        * records: The input tensor records.

        Returns:

        * tf.Tensor: The total number of tensor records.
        """
        dataset = records.map(lambda x, y: tf.constant(1, dtype=tf.int64), num_parallel_calls=tf.data.AUTOTUNE)
        # ignores any error encountered while reading the records
        # works with v2.x
        dataset = dataset.apply(tf.data.experimental.ignore_errors())
        n_tensors = dataset.reduce(np.int64(0), lambda x, y: x + y).numpy()
        return n_tensors

    @staticmethod
    def calculate_n_samples(**config):
        """
        Calculate the number of samples in the training, testing, and validation datasets.

        Parameters:

        **config: The configuration settings.

        Returns:

        * int: The number of training samples.

        * int: The number of testing samples.

        * int: The number of validation samples.

        """
        parser_tupler = partial(DataProcessor.parse_tfrecord,
                         patch_size=config.get("PATCH_SHAPE_SINGLE"),
                         features=config.get("FEATURES"),
                         labels=config.get("LABELS"),
                         x_only=False,
                         depth=config.get("OUT_CLASS_NUM"))

        tf_training_records = tf.data.Dataset.list_files(f"{str(config.get('TRAINING_DIR'))}/*")\
                                             .interleave(DataProcessor.create_tfrecord_from_file, num_parallel_calls=tf.data.AUTOTUNE)
        tf_training_records = tf_training_records.map(parser_tupler, num_parallel_calls=tf.data.AUTOTUNE)


        if config.get("PRINT_DATASET", False):
            DataProcessor.print_dataset_info(tf_training_records, "Training")

        n_training_records = DataProcessor.get_sum_tensor(tf_training_records)

        tf_testing_records = tf.data.Dataset.list_files(f"{str(config.get('TESTING_DIR'))}/*")\
                                            .interleave(DataProcessor.create_tfrecord_from_file, num_parallel_calls=tf.data.AUTOTUNE)
        tf_testing_records = tf_testing_records.map(parser_tupler, num_parallel_calls=tf.data.AUTOTUNE)
        n_testing_records = DataProcessor.get_sum_tensor(tf_testing_records)

        tf_validation_records = tf.data.Dataset.list_files(f"{str(config.get('VALIDATION_DIR'))}/*")\
                                               .interleave(DataProcessor.create_tfrecord_from_file, num_parallel_calls=tf.data.AUTOTUNE)
        tf_validation_records = tf_validation_records.map(parser_tupler, num_parallel_calls=tf.data.AUTOTUNE)
        n_validation_records = DataProcessor.get_sum_tensor(tf_validation_records)

        return n_training_records, n_testing_records, n_validation_records

    @staticmethod
    def print_dataset_info(dataset: tf.data.Dataset, dataset_name: str) -> None:
        """
        Print information about a dataset.

        Parameters:

        * dataset (tf.data.Dataset): The dataset to print information about.

        * dataset_name (str): The name of the dataset.
        """
        logging.info(dataset_name)
        for inputs, outputs in dataset.take(1):
            try:
                logging.info(f"inputs: {inputs.dtype.name} {inputs.shape}")
                logging.info(inputs)
                logging.info(f"outputs: {outputs.dtype.name} {outputs.shape}")
                logging.info(outputs)
            except:
                logging.info(f" > inputs:")
                for name, values in inputs.items():
                    logging.info(f"    {name}: {values.dtype.name} {values.shape}")
                # logging.info(f"    example \n: {dataset.take(1)}")
                logging.info(f" > outputs: {outputs.dtype.name} {outputs.shape}")

    @staticmethod
    @tf.function
    def random_transform(dataset: tf.Tensor, label: tf.Tensor) -> tf.Tensor:
        """
        Apply random transformations to a dataset.

        Parameters:

        * dataset (tf.Tensor): The input dataset.

        Returns:

        * tf.Tensor: The transformed dataset.
        """
        x = tf.random.uniform(())
        if x < 0.10:
            dataset = tf.image.flip_left_right(dataset)
            label = tf.image.flip_left_right(label)
        elif tf.math.logical_and(x >= 0.10, x < 0.20):
            dataset = tf.image.flip_up_down(dataset)
            label = tf.image.flip_up_down(label)
        elif tf.math.logical_and(x >= 0.20, x < 0.30):
            dataset = tf.image.flip_left_right(tf.image.flip_up_down(dataset))
            label = tf.image.flip_left_right(tf.image.flip_up_down(label))
        elif tf.math.logical_and(x >= 0.30, x < 0.40):
            dataset = tf.image.rot90(dataset, k=1)
            label = tf.image.rot90(label, k=1)
        elif tf.math.logical_and(x >= 0.40, x < 0.50):
            dataset = tf.image.rot90(dataset, k=2)
            label = tf.image.rot90(label, k=2)
        elif tf.math.logical_and(x >= 0.50, x < 0.60):
            dataset = tf.image.rot90(dataset, k=3)
            label = tf.image.rot90(label, k=3)
        elif tf.math.logical_and(x >= 0.60, x < 0.70):
            dataset = tf.image.flip_left_right(tf.image.rot90(dataset, k=2))
            label = tf.image.flip_left_right(tf.image.rot90(label, k=2))
        else:
            dataset = dataset
            label = label

        return dataset, label

    @staticmethod
    @tf.function
    def parse_tfrecord(example_proto: tf.Tensor, patch_size: int, features: list = None, labels: list = None, depth: int = 1) -> tf.data.Dataset:
        """
        Parse a TFRecord example.

        Parameters:
        * example_proto (tf.Tensor): The example to parse.

        * patch_size (int): The size of the patch.

        * features (list, optional): The list of feature names to include. Default is None.

        * labels (list, optional): The list of label names to include. Default is None.

        Returns:

        * tf.data.Dataset: The parsed dataset.
        """
        keys = features + labels
        columns = [
            tf.io.FixedLenFeature(shape=[patch_size, patch_size], dtype=tf.float32) for _ in keys
        ]
        proto_struct = dict(zip(keys, columns))
        inputs = tf.io.parse_single_example(example_proto, proto_struct)
        inputs_list = [inputs.get(key) for key in keys]
        stacked = tf.stack(inputs_list, axis=0)
        stacked = tf.transpose(stacked, [1, 2, 0])
        label = stacked[:, :, len(features):]
        y = tf.one_hot(tf.cast(label[:, :, -1], tf.uint8), depth)
        return stacked[:, :, :len(features)], y

    @staticmethod
    @tf.function
    def to_tuple(dataset: tf.Tensor, n_features: int = None, inverse_labels: bool = False) -> tuple:
        """
        Convert a dataset to a tuple of features and labels.

        Parameters:

        * dataset (tf.Tensor): The input dataset.

        * n_features (int, optional): The number of features. Default is None.

        * inverse_labels (bool, optional): Whether to inverse the labels. Default is False.

        Returns:

        * tuple: A tuple containing the features and labels.
        """
        features = dataset[:, :, :, :n_features]
        labels = dataset[:, :, :, n_features:]
        if inverse_labels:
            labels_inverse = tf.math.abs(labels - 1)
            labels = tf.concat([labels_inverse, labels], axis=-1)
        return features, labels

    @staticmethod
    @tf.function
    def parse_tfrecord_with_name(example_proto: tf.Tensor, patch_size: int, features: list = None, labels: list = None) -> tf.data.Dataset:
        """
        Parse a TFRecord example with named features.

        Parameters:

        * example_proto (tf.Tensor): The example to parse.

        * patch_size (int): The size of the patch.

        * features (list, optional): The list of feature names to include. Default is None.

        * labels (list, optional): The list of label names to include. Default is None.

        Returns:

        * tf.data.Dataset: The parsed dataset.
        """
        keys = features + labels
        columns = [
            tf.io.FixedLenFeature(shape=[patch_size, patch_size], dtype=tf.float32) for _ in keys
        ]
        proto_struct = dict(zip(keys, columns))
        return tf.io.parse_single_example(example_proto, proto_struct)

    @staticmethod
    @tf.function
    def to_tuple_with_name(inputs: tf.Tensor, features: list = None, labels: list = None, n_classes: int = 1) -> tuple:
        """
        Convert inputs with named features to a tuple of features and one-hot encoded labels.

        Parameters:

        * inputs (tf.Tensor): The input dataset.

        * features (list, optional): The list of feature names. Default is None.

        * labels (list, optional): The list of label names. Default is None.

        * n_classes (int, optional): The number of classes for one-hot encoding. Default is 1.

        Returns:

        * tuple: A tuple containing the features and one-hot encoded labels.
        """
        return (
            {name: inputs[name] for name in features},
            tf.one_hot(tf.cast(inputs[labels[0]], tf.uint8), n_classes)
        )

    @staticmethod
    @tf.function
    def parse_tfrecord_dnn(example_proto: tf.Tensor, features: list = None, labels: list = None) -> tuple:
        """
        Parse a TFRecord example for DNN models.

        Parameters:

        * example_proto (tf.Tensor): The example to parse.

        * features (list, optional): The list of feature names to include. Default is None.

        * labels (list, optional): The list of label names to include. Default is None.

        Returns:

        * tuple: A tuple containing the parsed features and labels.
        """
        keys = features + labels
        columns = [
            tf.io.FixedLenFeature(shape=[1], dtype=tf.float32) for _ in keys
        ]
        proto_struct = dict(zip(keys, columns))
        parsed_features = tf.io.parse_single_example(example_proto, proto_struct)
        label = parsed_features.pop(labels[0])
        label = tf.cast(label, tf.int32)
        return parsed_features, label

    @staticmethod
    @tf.function
    def to_tuple_dnn(dataset: dict, label: tf.Tensor, depth: int = 1) -> tuple:
        """
        Convert a dataset for DNN models to a tuple of features and one-hot encoded labels.

        Parameters:

        * dataset (dict): The input dataset.

        * label (tf.Tensor): The label.

        * depth (int, optional): The depth of one-hot encoding. Default is 1.

        Returns:

        * tuple: A tuple containing the features and one-hot encoded labels.
        """
        return tf.transpose(list(dataset.values())), tf.one_hot(indices=label, depth=depth)

    @staticmethod
    def to_tuple_dnn_ai_platform(dataset: dict, label: tf.Tensor, depth: int = 1) -> tuple:
        """
        Convert a dataset for DNN models to a tuple of features and one-hot encoded labels.

        Parameters:

        * dataset (dict): The input dataset.

        * label (tf.Tensor): The label.

        * depth (int, optional): The depth of one-hot encoding. Default is 1.

        Returns:

        * tuple: A tuple containing the features and one-hot encoded labels.
        """
          # (1) -> (1, 1, 1)
        return ({k: [[v]] for k, v in dataset.items()}, tf.expand_dims(tf.one_hot(label, depth), axis=0))


    @staticmethod
    @tf.function
    def parse_tfrecord_multi_label(example_proto: tf.data.Dataset, patch_size: int, features: list = None, labels: list = None) -> tuple:
        """
        Parse a TFRecord example with multiple labels.

        Parameters:

        * example_proto (tf.data.Dataset): The example to parse.

        * patch_size (int): The size of the patch.

        * features (list, optional): The list of feature names to include. Default is None.

        * labels (list, optional): The list of label names to include. Default is None.

        Returns:

        * tuple: A tuple containing the parsed features and labels.
        """
        keys = features + labels
        columns = [
            tf.io.FixedLenFeature(shape=[patch_size, patch_size], dtype=tf.float32) for _ in keys
        ]
        proto_struct = dict(zip(keys, columns))
        parsed_features = tf.io.parse_single_example(example_proto, proto_struct)
        label = parsed_features.pop(labels[0])
        return parsed_features, label

    @staticmethod
    @tf.function
    def to_tuple_multi_label(dataset: dict, label: tf.Tensor, depth: int = 1, x_only: bool = False) -> tuple:
        """
        Convert a dataset with multiple labels to a tuple of features and multi-hot encoded labels.

        Parameters:

        * dataset (tuple): The input dataset.

        * n_labels (int, optional): The number of labels. Default is 1.

        Returns:

        * tuple: A tuple containing the features and multi-hot encoded labels.
        """
        label = tf.cast(label, tf.uint8)
        label = tf.one_hot(indices=label, depth=depth)
        parsed_dataset = {k: tf.expand_dims(v, axis=2) for k, v in dataset.items()}
        if x_only:
            return parsed_dataset
        return parsed_dataset, label


    @staticmethod
    @tf.function
    def to_tuple_multi_label_ai_platform(dataset: dict, label: tf.Tensor, depth: int = 1) -> tuple:
        """
        Convert a dataset with multiple labels to a tuple of features and multi-hot encoded labels.

        Parameters:

        * dataset (tuple): The input dataset.

        * n_labels (int, optional): The number of labels. Default is 1.

        Returns:

        * tuple: A tuple containing the features and multi-hot encoded labels.
        """
        label = tf.cast(label, tf.uint8)
        label = tf.one_hot(indices=label, depth=depth)
        parsed_dataset = {k: tf.expand_dims(v, axis=2) for k, v in dataset.items()}
        return parsed_dataset, label


    @staticmethod
    def _get_dataset(files: list, features: list, labels: list, patch_shape: list, batch_size: int, buffer_size: int = 1000, training: bool = False, **kwargs) -> tf.data.Dataset:
        """
        Get a TFRecord dataset.

        Parameters:
        filenames (list): The list of file names.
        patch_size (int): The size of the patch.
        features (list, optional): The list of feature names to include. Default is None.
        labels (list, optional): The list of label names to include. Default is None.
        batch_size (int, optional): The batch size. Default is 1.
        shuffle (bool, optional): Whether to shuffle the dataset. Default is False.
        n_labels (int, optional): The number of labels. Default is 1.
        num_parallel_calls (int, optional): The number of parallel calls. Default is tf.data.experimental.AUTOTUNE.
        drop_remainder (bool, optional): Whether to drop the remainder of batches. Default is False.
        cache (bool, optional): Whether to cache the dataset. Default is False.

        Returns:

        tf.data.Dataset: The TFRecord dataset.
        """
        dnn = kwargs.get('dnn', False)
        inverse_labels = kwargs.get('inverse_labels', False)
        depth = kwargs.get('depth', len(labels))
        multi_label_unet = kwargs.get('multi_label_unet', False)
        dataset = tf.data.TFRecordDataset(files, compression_type='GZIP')

        if dnn:
            parser = partial(DataProcessor.parse_tfrecord_dnn, features=features, labels=labels)
            split_data = partial(DataProcessor.to_tuple_dnn, depth=depth)
            dataset = dataset.map(parser, num_parallel_calls=tf.data.experimental.AUTOTUNE)
            dataset = dataset.map(split_data, num_parallel_calls=tf.data.experimental.AUTOTUNE)
            dataset = dataset.batch(batch_size)
            return dataset

        if multi_label_unet:
            parser = partial(DataProcessor.parse_tfrecord_multi_label, features=features, labels=labels, patch_shape=patch_shape)
            split_data = partial(DataProcessor.to_tuple_multi_label, n_features=len(features), depth=depth)
            dataset = dataset.interleave(parser, num_parallel_calls=tf.data.experimental.AUTOTUNE)
            if training:
                dataset = dataset.shuffle(buffer_size, reshuffle_each_iteration=True).batch(batch_size) \
                                 .map(DataProcessor.random_transform, num_parallel_calls=tf.data.experimental.AUTOTUNE) \
                                 .map(split_data, num_parallel_calls=tf.data.experimental.AUTOTUNE)
            else:
                dataset = dataset.batch(batch_size).map(split_data, num_parallel_calls=tf.data.experimental.AUTOTUNE)

            return dataset

        parser = partial(DataProcessor.parse_tfrecord, features=features, labels=labels)
        split_data = partial(DataProcessor.to_tuple, n_features=len(features), inverse_labels=inverse_labels)
        dataset = dataset.interleave(parser, num_parallel_calls=tf.data.experimental.AUTOTUNE)

        if training:
            dataset = dataset.shuffle(buffer_size, reshuffle_each_iteration=True).batch(batch_size) \
                             .map(DataProcessor.random_transform, num_parallel_calls=tf.data.experimental.AUTOTUNE) \
                             .map(split_data, num_parallel_calls=tf.data.experimental.AUTOTUNE)
        else:
            dataset = dataset.batch(batch_size).map(split_data, num_parallel_calls=tf.data.experimental.AUTOTUNE)

        return dataset

    @staticmethod
    def get_dataset(pattern: str, features: list, labels: list, patch_size: int, batch_size: int, n_classes: int = 1, **kwargs) -> tf.data.Dataset:
        """
        Get a TFRecord dataset.

        Parameters:
        * filenames (list): The list of file names.

        * patch_size (int): The size of the patch.

        * (list, optional): The list of feature names to include. Default is None.

        * labels (list, optional): The list of label names to include. Default is None.

        * batch_size (int, optional): The batch size. Default is 1.

        * shuffle(bool, optional): Whether to shuffle the dataset. Default is False.

        * n_labels (int, optional): The number of labels. Default is 1.

        * num_parallel_calls (int, optional): The number of parallel calls. Default is tf.data.experimental.AUTOTUNE.

        * drop_remainder (bool, optional): Whether to drop the remainder of batches. Default is False.

        * cache (bool, optional): Whether to cache the dataset. Default is False.

        Returns:
        * tf.data.Dataset: The TFRecord dataset.
        """
        logging.info(f"Loading dataset from {pattern}")

        dataset = tf.data.Dataset.list_files(pattern).interleave(DataProcessor.create_tfrecord_from_file)

        if kwargs.get("IS_DNN", False):
            if kwargs.get("USE_AI_PLATFORM", False):
                parser = partial(DataProcessor.parse_tfrecord_dnn, features=features, labels=labels)
                tupler = partial(DataProcessor.to_tuple_dnn_ai_platform, depth=n_classes)
            else:
                parser = partial(DataProcessor.parse_tfrecord_dnn, features=features, labels=labels)
                tupler = partial(DataProcessor.to_tuple_dnn, depth=n_classes)

            dataset = dataset.map(parser, num_parallel_calls=tf.data.AUTOTUNE)
            dataset = dataset.map(tupler, num_parallel_calls=tf.data.AUTOTUNE)
            dataset = dataset.shuffle(512)
            dataset = dataset.batch(batch_size)
            dataset = dataset.prefetch(buffer_size=tf.data.AUTOTUNE)
            return dataset

        if kwargs.get("USE_AI_PLATFORM", False):
            parser = partial(DataProcessor.parse_tfrecord_multi_label, patch_size=patch_size, features=features, labels=labels)
            tupler = partial(DataProcessor.to_tuple_multi_label_ai_platform, depth=n_classes)
            parser_tupler = None
        else:
            parser_tupler = partial(DataProcessor.parse_tfrecord, patch_size=patch_size, features=features, labels=labels, depth=n_classes)

        if parser_tupler is not None:
            dataset = dataset.map(parser_tupler, num_parallel_calls=tf.data.AUTOTUNE)
        else:
            dataset = dataset.map(parser, num_parallel_calls=tf.data.AUTOTUNE)
            dataset = dataset.map(tupler, num_parallel_calls=tf.data.AUTOTUNE)

        dataset = dataset.shuffle(512)
        # dataset = dataset.batch(batch_size)
        dataset = dataset.prefetch(buffer_size=tf.data.AUTOTUNE)
        if kwargs.get("training", False) and kwargs.get("TRANSFORM_DATA", True):
            logging.info("randomly transforming data")
            if kwargs.get("USE_AI_PLATFORM", False):
                dataset = dataset.map(RandomTransform(), num_parallel_calls=tf.data.AUTOTUNE)
            else:
                dataset = dataset.map(DataProcessor.random_transform, num_parallel_calls=tf.data.AUTOTUNE)

        dataset = dataset.batch(batch_size)
        # dataset = dataset.cache()
        return dataset



class RandomTransform(tf.keras.layers.Layer):
    def __init__(self, seed=42, unit_range=True):
        super().__init__()
        self.seed = seed
        self.flip_horizontal = tf.keras.layers.RandomFlip("horizontal", seed=self.seed)
        self.flip_vertical = tf.keras.layers.RandomFlip("vertical", seed=self.seed)
        self.flip_both = tf.keras.layers.RandomFlip("horizontal_and_vertical", seed=self.seed)
        self.random_brightness = tf.keras.layers.RandomBrightness(0.2, value_range=(0, 1) if unit_range else (0, 255), seed=self.seed)
        self.random_contrast = tf.keras.layers.RandomContrast(0.2, seed=self.seed)

    @tf.function
    def call(self, dataset, label):
        """
        Apply random transformations to a dataset.

        Parameters:

        * dataset (tf.Tensor): The input dataset.

        * label (tf.Tensor): The corresponding label.

        Returns:

        * tuple: The transformed dataset and label as a tuple.
        """
        x = tf.random.uniform((), seed=self.seed)
        transformed_features = {}

        # Apply the same random transformation across all bands or features
        for key, feature in dataset.items():
            transformed_feature = feature  # Default to no change
            if x < 0.10:
                transformed_feature = self.flip_horizontal(feature)
            elif tf.math.logical_and(x >= 0.10, x < 0.20):
                transformed_feature = self.flip_vertical(feature)
            elif tf.math.logical_and(x >= 0.20, x < 0.30):
                transformed_feature = self.flip_both(feature)
            elif tf.math.logical_and(x >= 0.30, x < 0.40):
                transformed_feature = tf.image.rot90(feature, k=1)
            elif tf.math.logical_and(x >= 0.40, x < 0.50):
                transformed_feature = tf.image.rot90(feature, k=2)
            elif tf.math.logical_and(x >= 0.50, x < 0.60):
                transformed_feature = tf.image.rot90(feature, k=3)
            elif tf.math.logical_and(x >= 0.60, x < 0.70):
                transformed_feature = self.random_brightness(feature)
            elif tf.math.logical_and(x >= 0.70, x < 0.80):
                transformed_feature = self.random_contrast(feature)

            transformed_features[key] = transformed_feature

        # Apply corresponding transformations to the label
        transformed_label = label  # Default to no change
        if x < 0.10:
            transformed_label = self.flip_horizontal(label)
        elif tf.math.logical_and(x >= 0.10, x < 0.20):
            transformed_label = self.flip_vertical(label)
        elif tf.math.logical_and(x >= 0.20, x < 0.30):
            transformed_label = self.flip_both(label)
        elif tf.math.logical_and(x >= 0.30, x < 0.40):
            transformed_label = tf.image.rot90(label, k=1)
        elif tf.math.logical_and(x >= 0.40, x < 0.50):
            transformed_label = tf.image.rot90(label, k=2)
        elif tf.math.logical_and(x >= 0.50, x < 0.60):
            transformed_label = tf.image.rot90(label, k=3)

        return transformed_features, transformed_label



# -*- coding: utf-8 -*-

"""
This script generates training data for a machine learning model.
The data is split into training, validation, and test sets using the Apache Beam library.
Earth Engine Python API is used to load and manipulate geospatial data.
The script uses TensorFlow to serialize the data into TFRecord format.
"""

import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions

import ee
import argparse

try:
    from aces.ee_utils import EEUtils
    from aces.utils import TFUtils, Utils
    from aces.config import Config
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))
    from aces.ee_utils import EEUtils
    from aces.utils import TFUtils, Utils
    from aces.config import Config

# Before running this script, you need to authenticate to Google Cloud:
# https://cloud.google.com/dataflow/docs/quickstarts/create-pipeline-python#before-you-begin

# ToDo: Is there a way to use seed value in the split?

__all__ = ["TrainingDataGenerator"]

class TrainingDataGenerator:
    """
    A class to generate training data for machine learning models.
    This class utilizes Apache Beam for data processing and Earth Engine for handling geospatial data.
    """
    def __init__(self, include_after: bool = False, use_service_account: bool = False):
        """
        Constructor for the TrainingDataGenerator class.

        Parameters:
        include_after (bool): If True, includes "after" images in the generated data. Default is False.
        """
        self.output_bucket = Config.GCS_BUCKET
        self.kernel_size = Config.PATCH_SHAPE_SINGLE
        self.grace = 10
        self.scale = Config.SCALE
        self.include_after = include_after
        self.test_ratio = 0.1
        self.validation_ratio = 0.2
        self.seed = 100
        self.use_service_account = use_service_account

    def load_data(self) -> None:
        """
        Load the necessary data from Earth Engine and prepare it for use.
        """
        EEUtils.initialize_session(use_highvolume=True, key=Config.EE_SERVICE_CREDENTIALS if self.use_service_account else None)
        self.l1 = ee.FeatureCollection("projects/servir-sco-assets/assets/Bhutan/BT_Admin_1")
        self.paro = self.l1.filter(ee.Filter.eq("ADM1_EN", "Paro"))

        # self.sample_locations = ee.FeatureCollection("projects/servir-sco-assets/assets/Bhutan/ACES_2/paro_2021_all_class_samples")
        # self.sample_locations = ee.FeatureCollection("projects/servir-sco-assets/assets/Bhutan/ACES_2/paro_2021_all_class_samples_clipped")
        self.sample_locations = ee.FeatureCollection("projects/servir-sco-assets/assets/Bhutan/ACES_2/paro_2021_all_class_samples_clipped_new")
        self.sample_locations = self.sample_locations.randomColumn("random", self.seed)
        self.training_sample_locations = self.sample_locations.filter(ee.Filter.gt("random", self.validation_ratio + self.test_ratio)) # > 0.4
        print("Training sample size:", self.training_sample_locations.size().getInfo())
        self.validation_sample_locations = self.sample_locations.filter(ee.Filter.lte("random", self.validation_ratio)) # <= 0.2
        print("Validation sample size:", self.validation_sample_locations.size().getInfo())
        self.test_sample_locations = self.sample_locations.filter(ee.Filter.And(ee.Filter.gt("random", self.validation_ratio),
                                                               ee.Filter.lte("random", self.validation_ratio + self.test_ratio))) # > 0.2 and <= 0.4
        print("Test sample size:", self.test_sample_locations.size().getInfo())
        self.sample_size = self.sample_locations.size().getInfo()
        print("Sample size:", self.sample_size)
        self.sample_locations_list = self.sample_locations.toList(self.sample_size + self.grace)

        self.label = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/paro_2021_all_class_label").rename("class").unmask(0, False)
        # self.other = self.label.remap([0, 1], [1, 0]).rename(["other"])

        self.composite_after = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/Paro_Rice_Composite_2021/composite_after")

        self.composite_before = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/Paro_Rice_Composite_2021/composite_before")

        self.composite_during = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/Paro_Rice_Composite_2021/composite_during")

        evi_before = EEUtils.calculate_evi(self.composite_before)
        evi_during = EEUtils.calculate_evi(self.composite_during)
        evi_after = EEUtils.calculate_evi(self.composite_after)

        self.composite_before = self.composite_before.addBands(evi_before)
        self.composite_during = self.composite_during.addBands(evi_during)
        self.composite_after = self.composite_after.addBands(evi_after)

        original_bands = self.composite_before.bandNames().getInfo()
        lowercase_bands = [band.lower() for band in original_bands]

        self.composite_before = self.composite_before.select(original_bands, lowercase_bands)
        self.composite_after = self.composite_after.select(original_bands, lowercase_bands)
        self.composite_during = self.composite_during.select(original_bands, lowercase_bands)

        self.composite_before = self.composite_before.regexpRename("$(.*)", "_before")
        self.composite_after = self.composite_after.regexpRename("$(.*)", "_after")
        self.composite_during = self.composite_during.regexpRename("$(.*)", "_during")

        self.sentinel1_asc_before_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Ascending2021/s1AscBefore")
        self.sentinel1_asc_during_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Ascending2021/s1AscDuring")

        self.sentinel1_desc_before_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Descending2021/s1DescBefore")
        self.sentinel1_desc_during_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Descending2021/s1DescDuring")

        self.elevation = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/elevationParo")
        self.slope = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/slopeParo")

        if self.include_after:
            self.image = self.composite_before.addBands(self.composite_during).addBands(self.composite_after).toFloat()
        else:
            self.image = self.composite_before.addBands(self.composite_during).toFloat()

        if Config.USE_S1:
            self.image = self.image.addBands(self.sentinel1_asc_before_composite).addBands(self.sentinel1_asc_during_composite).addBands(self.sentinel1_desc_before_composite).addBands(self.sentinel1_desc_during_composite).toFloat()
            Config.FEATURES.extend(["vv_asc_before", "vh_asc_before", "vv_asc_during", "vh_asc_during",
                                    "vv_desc_before", "vh_desc_before", "vv_desc_during", "vh_desc_during"])

        if Config.USE_ELEVATION:
            self.image = self.image.addBands(self.elevation).addBands(self.slope).toFloat()
            Config.FEATURES.extend(["elevation", "slope"])

        self.image = self.image.select(Config.FEATURES)
        self.image = self.image.addBands(self.label).toFloat()

        print("Image bands:", self.image.bandNames().getInfo())
        self.selectors = self.image.bandNames().getInfo()

        proj = ee.Projection("EPSG:4326").atScale(10).getInfo()

        # Get scales out of the transform.
        self.scale_x = proj["transform"][0]
        self.scale_y = -proj["transform"][4]

    def generate_training_neighborhood_data(self) -> None:
        """
        Use Apache Beam to generate training, validation, and test patch data from the loaded data.
        """
        from datetime import datetime
        from uuid import uuid4

        export_kwargs = { "bucket": self.output_bucket, "selectors": self.selectors }

        file_prefix = f"experiments_neighbour_{self.kernel_size}x{self.kernel_size}"

        if Config.USE_S1:
            file_prefix += "_s1"

        if Config.USE_ELEVATION:
            file_prefix += "_elevation"

        training_file_prefix = f"{file_prefix}_training/training"
        testing_file_prefix = f"{file_prefix}_testing/testing"
        validation_file_prefix = f"{file_prefix}_validation/validation"

        def _generate_data(image, data, selectors, use_service_account, prefix):
            if "training" in prefix:
                description = "Training"
            elif "validation" in prefix:
                description = "Validation"
            elif "testing" in prefix:
                description = "Testing"
            else:
                description = "Unknown"

            print(f"{description} Data")

            beam_options = PipelineOptions([], direct_num_workers=0, direct_running_mode="multi_processing", runner="DirectRunner")
            with beam.Pipeline(options=beam_options) as pipeline:
                _ = (
                    pipeline
                    | "Create range" >> beam.Create(range(0, data.size().getInfo(), 1))
                    | "Yield sample points" >> beam.Map(EEUtils.beam_yield_sample_points_with_index, data.toList(data.size()), use_service_account)
                    | "Get patch" >> beam.Map(EEUtils.beam_sample_neighbourhood, image, use_service_account)
                    | "Write training data" >> beam.Map(EEUtils.beam_export_collection_to_cloud_storage, start_training=True,
                                                        **{**export_kwargs, "file_prefix": f"{prefix}_{datetime.now().strftime('%Y%m-%d%H-%M-%S_')}",
                                                           "description": f"{description}_{datetime.now().strftime('%Y%m-%d%H-%M-%S_')}",
                                                           "selectors": selectors})
                )

        _generate_data(self.image, self.training_sample_locations, self.selectors, self.use_service_account, training_file_prefix)

        _generate_data(self.image, self.validation_sample_locations, self.selectors, self.use_service_account, validation_file_prefix)

        _generate_data(self.image, self.test_sample_locations, self.selectors, self.use_service_account, testing_file_prefix)

    def generate_training_patch_data(self) -> None:
        """
        Use Apache Beam to generate training, validation, and test patch data from the loaded data.
        """
        beam_options = PipelineOptions([], direct_num_workers=0, direct_running_mode="multi_processing", runner="DirectRunner")
        with beam.Pipeline(options=beam_options) as pipeline:
            training_data, validation_data, test_data = (
                pipeline
                | "Create range" >> beam.Create(range(0, self.sample_size, 1))
                | "Yield sample points" >> beam.Map(EEUtils.beam_yield_sample_points, self.sample_locations_list, self.use_service_account)
                | "Get patch" >> beam.Map(EEUtils.beam_get_training_patches, self.image, self.selectors, self.scale, self.kernel_size, self.use_service_account)
                | "Filter patches" >> beam.Filter(Utils.filter_good_patches)
                | "Serialize" >> beam.Map(TFUtils.beam_serialize)
                | "Split dataset" >> beam.Partition(Utils.split_dataset, 3, validation_ratio=self.validation_ratio, test_ratio=self.test_ratio)
            )

            # Write the datasets to TFRecord files in the output bucket
            training_data | "Write training data" >> beam.io.WriteToTFRecord(
                f"gs://{self.output_bucket}/experiments_paro_{self.kernel_size}x{self.kernel_size}_before_during{'_after' if self.include_after else ''}_training/training", file_name_suffix=".tfrecord.gz"
            )
            validation_data | "Write validation data" >> beam.io.WriteToTFRecord(
                f"gs://{self.output_bucket}/experiments_paro_{self.kernel_size}x{self.kernel_size}_before_during{'_after' if self.include_after else ''}_validation/validation", file_name_suffix=".tfrecord.gz"
            )
            test_data | "Write test data" >> beam.io.WriteToTFRecord(
                f"gs://{self.output_bucket}/experiments_paro_{self.kernel_size}x{self.kernel_size}_before_during{'_after' if self.include_after else ''}_testing/testing", file_name_suffix=".tfrecord.gz"
            )

    def generate_training_patch_seed_data(self) -> None:
        """
        Use Apache Beam to generate training, validation, and test patch data from the loaded data.
        """
        def _generate_data_seed(image, data, selectors, scale, kernel_size, use_service_account, output_path) -> None:
            beam_options = PipelineOptions([], direct_num_workers=0, direct_running_mode="multi_processing", runner="DirectRunner")
            with beam.Pipeline(options=beam_options) as pipeline:
                _ = (
                    pipeline
                    | "Create range" >> beam.Create(range(0, data.size().getInfo(), 1))
                    | "Yield sample points" >> beam.Map(EEUtils.beam_yield_sample_points, data.toList(data.size()), use_service_account)
                    | "Get patch" >> beam.Map(EEUtils.beam_get_training_patches, image, selectors, scale, kernel_size, use_service_account)
                    | "Filter patches" >> beam.Filter(Utils.filter_good_patches)
                    | "Serialize" >> beam.Map(TFUtils.beam_serialize)
                    | "Write training data" >> beam.io.WriteToTFRecord(output_path, file_name_suffix=".tfrecord.gz")
                )

        output_path = f"gs://{self.output_bucket}/unet_{self.kernel_size}x{self.kernel_size}_planet_wo_indices"

        if Config.USE_S1:
            output_path += "_w_s1"

        if Config.USE_ELEVATION:
            output_path += "_w_elevation"

        datasets = [
            {"name": "training", "locations": self.training_sample_locations, "output_path": f"{output_path}/training_70/training"},
            {"name": "testing", "locations": self.test_sample_locations, "output_path": f"{output_path}/testing_10/testing"},
            {"name": "validation", "locations": self.validation_sample_locations, "output_path": f"{output_path}/validation_20/validation"}
        ]

        for dataset in datasets:
            print(f"{dataset['name'].capitalize()} output path:", dataset["output_path"])
            _generate_data_seed(
                self.image, dataset["locations"], self.selectors, self.scale, self.kernel_size,
                self.use_service_account, dataset["output_path"]
            )

    def generate_training_point_data(self) -> None:
        """
        Use Apache Beam to generate training, validation, and test point data from the loaded data.
        """
        training_sample_points = EEUtils.sample_image(self.image, self.training_sample_locations, **Config.__dict__)
        validation_sample_points = EEUtils.sample_image(self.image, self.validation_sample_locations, **Config.__dict__)
        test_sample_points = EEUtils.sample_image(self.image, self.test_sample_locations, **Config.__dict__)

        output_path = f"dnn_planet_wo_indices"

        if Config.USE_S1:
            output_path += "_w_s1"

        if Config.USE_ELEVATION:
            output_path += "_w_elevation"

        datasets = [
            {"name": "training", "sample_points": training_sample_points, "output_path": f"{output_path}/training_70/training"},
            {"name": "testing", "sample_points": test_sample_points, "output_path": f"{output_path}/testing_10/testing"},
            {"name": "validation", "sample_points": validation_sample_points, "output_path": f"{output_path}/validation_20/validation"}
        ]

        export_kwargs = { "bucket": self.output_bucket, "selectors": self.selectors }

        for dataset in datasets:
            print(f"{dataset['name'].capitalize()} output path:", dataset["output_path"])
            EEUtils.export_collection_data(dataset["sample_points"], export_type="cloud", start_training=True, **{**export_kwargs, "file_prefix": dataset["output_path"], "description": dataset["name"].capitalize()})

    def run_patch_generator(self) -> None:
        """
        Run the patch training data generation process.
        """
        self.load_data()
        self.generate_training_patch_data()

    def run_patch_generator_seed(self) -> None:
        """
        Run the patch training data generation process.
        """
        self.load_data()
        self.generate_training_patch_seed_data()

    def run_point_generator(self) -> None:
        """
        Run the point training data generation process.
        """
        self.load_data()
        self.generate_training_point_data()

    def run_neighborhood_generator(self) -> None:
        """
        Run the neighborhood training data generation process.
        """
        self.load_data()
        self.generate_training_neighborhood_data()


if __name__ == "__main__":
    print("Program started..")

    # Initialize parser
    parser = argparse.ArgumentParser()

    # Adding optional argument
    parser.add_argument("-m", "--mode", help="""Which mode to run? The available modes are: patch, point, neighborhood, patch_seed. \n
                        use as python generate_training_data.py --mode patch \n
                        or python generate_training_data.py --m point""")

    # Read arguments from command line
    mode = "neighborhood"
    if parser.parse_args().mode:
        mode = parser.parse_args().mode
    else:
        print("No mode specified, defaulting to neighborhood.")

    generator = TrainingDataGenerator(use_service_account=Config.USE_SERVICE_ACCOUNT)
    if mode == "patch":
        generator.run_patch_generator()
    elif mode == "patch_seed":
        generator.run_patch_generator_seed()
    elif mode == "point":
        generator.run_point_generator()
    elif mode == "neighborhood":
        generator.run_neighborhood_generator()
    else:
        print("Invalid mode specified.")

    print("\nProgram completed.")



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)

try:
    from aces.config import Config
except ModuleNotFoundError:
    logging.info("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

    from aces.config import Config

import os
import tensorflow as tf
import subprocess


# Load the trained model
# h5 won't work with vertex ai
# this_model = tf.keras.models.load_model(f"{str(Config.MODEL_DIR)}/{Config.MODEL_NAME}.h5")

if Config.USE_BEST_MODEL_FOR_INFERENCE:
    logging.info(f"Using best model for inference.\nLoading model from {str(Config.MODEL_DIR)}/{Config.MODEL_CHECKPOINT_NAME}.tf")
    this_model = tf.keras.models.load_model(f"{str(Config.MODEL_DIR)}/{Config.MODEL_CHECKPOINT_NAME}")
else:
    logging.info(f"Using last model for inference.\nLoading model from {str(Config.MODEL_DIR)}/{Config.MODEL_NAME}")
    this_model = tf.keras.models.load_model(f"{str(Config.MODEL_DIR)}/{Config.MODEL_NAME}")

logging.info(this_model.summary())

# Save the trained model in the google cloud storage bucket
gcs_save_dir = f"gs://{Config.GCS_BUCKET}/{Config.GCS_VERTEX_MODEL_SAVE_DIR}/{Config.MODEL_DIR_NAME}"
this_model.save(gcs_save_dir)
logging.info(f"Model saved to: {gcs_save_dir}")

def run_shell_process(exe):
    p = subprocess.Popen(exe, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    while(True):
        # returns None while subprocess is running
        retcode = p.poll()
        line = p.stdout.readline()
        yield str(line)
        if retcode is not None:
            break

Config.MODEL_DIR_NAME = f"{Config.MODEL_DIR_NAME}_mt_{Config.GCP_MACHINE_TYPE.replace('-', '_')}"
Config.MODEL_DIR_NAME = "dnn_w_elv"

# Deploy
# delete model before deploying
try:
    delete_model = f"gcloud ai models delete {Config.MODEL_DIR_NAME} --project={Config.GCS_PROJECT} --region={Config.GCS_REGION}"
    logging.info(f"deleting model before deploying: {delete_model}")
    for line in run_shell_process(delete_model.split()):
        logging.info(line)
except Exception as e:
    logging.info(f"Exception in deleting model: {e}")

# upload model
upload_model = f"""gcloud ai models upload \
--artifact-uri={gcs_save_dir} \
--project={Config.GCS_PROJECT} \
--region={Config.GCS_REGION} \
--container-image-uri={Config.GCS_VERTEX_CONTAINER_IMAGE} \
--description={Config.MODEL_DIR_NAME} \
--display-name={Config.MODEL_DIR_NAME} \
--model-id={Config.MODEL_DIR_NAME}"""

logging.info(f"uploading model")
result = subprocess.check_output(upload_model, shell=True)
logging.info(f"{result}")

# deploy model

# create end_point
endpoint_create = f"""gcloud ai endpoints create \
--display-name={Config.MODEL_DIR_NAME + '_end_point'} \
--region={Config.GCS_REGION} \
--project={Config.GCS_PROJECT}"""

logging.info(f"creating end point")
result = subprocess.check_output(endpoint_create, shell=True)
logging.info(f"{result}")

# get end_point
endpoint_get = f"""gcloud ai endpoints list \
--project={Config.GCS_PROJECT} \
--region={Config.GCS_REGION} \
--filter=displayName:{Config.MODEL_DIR_NAME} \
--format="value(ENDPOINT_ID.scope())"\
"""
logging.info(f"getting endpoint")
result = subprocess.check_output(endpoint_get, shell=True)
logging.info(f"{str(result)}")

endpoints = result.decode("utf-8").split()
endpoint_id = endpoints[0]

# deploy the model
# deploy_model = f"""gcloud ai endpoints deploy-model {endpoint_id} \
# --project={Config.GCS_PROJECT} \
# --region={Config.GCS_REGION} \
# --model={Config.MODEL_DIR_NAME} \
# --display-name={Config.MODEL_DIR_NAME} \
# --machine-type={Config.GCP_MACHINE_TYPE}"""
# logging.info(f"deploying model:")
# result = subprocess.check_output(deploy_model, shell=True)
# logging.info(result)


deploy_model = f"""gcloud ai endpoints deploy-model {endpoint_id} \
--project={Config.GCS_PROJECT} \
--region={Config.GCS_REGION} \
--model={Config.MODEL_DIR_NAME} \
--display-name={Config.MODEL_DIR_NAME} \
--machine-type={Config.GCP_MACHINE_TYPE}"""
# --accelerator=type=nvidia-tesla-k80,count=1"""
# --deployed-model-id={Config.MODEL_DIR_NAME}"""
# --traffic-split=[{Config.MODEL_DIR_NAME}=100]"""
logging.info(f"deploying model:")
logging.info(deploy_model)
result = subprocess.check_output(deploy_model, shell=True)
logging.info(result)



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)

try:
    from aces.config import Config
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

    from aces.config import Config

import json
import os
import tensorflow as tf
import numpy as np
import subprocess


OUTPUT_IMAGE_FILE = str(Config.MODEL_DIR / "prediction" / f"{Config.OUTPUT_NAME}.TFRecord")
if not os.path.exists(str(Config.MODEL_DIR / "prediction")): os.mkdir(str(Config.MODEL_DIR / "prediction"))
print(f"OUTPUT_IMAGE_FILE: {OUTPUT_IMAGE_FILE}")

OUTPUT_GCS_PATH = f"gs://{Config.GCS_BUCKET}/prediction/{Config.OUTPUT_NAME}.TFRecord"
print(f"OUTPUT_GCS_PATH: {OUTPUT_GCS_PATH}")

ls = f"sudo gsutil ls gs://{Config.GCS_BUCKET}/{Config.GCS_IMAGE_DIR}/"
print(f"ls >> : {ls}")
files_list = subprocess.check_output(ls, shell=True)
files_list = files_list.decode("utf-8")
files_list = files_list.split("\n")

# Get only the files generated by the image export.
exported_files_list = [s for s in files_list if Config.GCS_IMAGE_PREFIX in s]

print(f"exported_files_list: {exported_files_list}")

# Get the list of image files and the JSON mixer file.
image_files_list = []
json_file = None
for f in exported_files_list:
    if f.endswith(".tfrecord.gz"):
        image_files_list.append(f)
    elif f.endswith(".json"):
        json_file = f

# Make sure the files are in the right order.
image_files_list.sort()

print(f"image_files_list: {image_files_list}")

print(f"json_file: {json_file}")

if Config.USE_BEST_MODEL_FOR_INFERENCE:
    logging.info(f"Using best model for inference.\nLoading model from {str(Config.MODEL_DIR)}/{Config.MODEL_CHECKPOINT_NAME}.tf")
    this_model = tf.keras.models.load_model(f"{str(Config.MODEL_DIR)}/{Config.MODEL_CHECKPOINT_NAME}.tf")
else:
    logging.info(f"Using last model for inference.\nLoading model from {str(Config.MODEL_DIR)}/trained-model")
    this_model = tf.keras.models.load_model(f"{str(Config.MODEL_DIR)}/trained-model")

logging.info(this_model.summary())


cat = f"gsutil cat {json_file}"
read_t = subprocess.check_output(cat, shell=True)
read_t = read_t.decode("utf-8")

# Get a single string w/ newlines from the IPython.utils.text.SList
mixer = json.loads(read_t)

# Get relevant info from the JSON mixer file.
patch_width = mixer["patchDimensions"][0]
patch_height = mixer["patchDimensions"][1]
patches = mixer["totalPatches"]
patch_dimensions_flat = [patch_width * patch_height, 1]

# Get set up for prediction.
if Config.KERNEL_BUFFER:
    x_buffer = Config.KERNEL_BUFFER[0] // 2
    y_buffer = Config.KERNEL_BUFFER[1] // 2

    buffered_shape = [
        Config.PATCH_SHAPE[0] + Config.KERNEL_BUFFER[0],
        Config.PATCH_SHAPE[1] + Config.KERNEL_BUFFER[1],
    ]
else:
    x_buffer = 0
    y_buffer = 0
    buffered_shape = Config.PATCH_SHAPE

if Config.USE_ELEVATION:
    Config.FEATURES.extend(["elevation", "slope"])


if Config.USE_S1:
    Config.FEATURES.extend(["vv_asc_before", "vh_asc_before", "vv_asc_during", "vh_asc_during",
                            "vv_desc_before", "vh_desc_before", "vv_desc_during", "vh_desc_during"])

print(f"Config.FEATURES: {Config.FEATURES}")

image_columns = [
    tf.io.FixedLenFeature(shape=buffered_shape, dtype=tf.float32) for k in Config.FEATURES
]

image_features_dict = dict(zip(Config.FEATURES, image_columns))

def parse_image(example_proto):
    return tf.io.parse_single_example(example_proto, image_features_dict)

def toTupleImage(inputs):
    inputsList = [inputs.get(key) for key in Config.FEATURES]
    stacked = tf.stack(inputsList, axis=0)
    stacked = tf.transpose(stacked, [1, 2, 0])
    return stacked

# Create a dataset from the TFRecord file(s) in Cloud Storage.
image_dataset = tf.data.TFRecordDataset(image_files_list, compression_type="GZIP")
image_dataset = image_dataset.map(parse_image, num_parallel_calls=5)
image_dataset = image_dataset.map(toTupleImage).batch(1)

# for inputs in image_dataset.take(1):
#     print("inputs", inputs)
#     print(f"inputs: {inputs.shape}")

# Perform inference.
print("Running predictions...")
print(f"patches: {patches}")
predictions = this_model.predict(image_dataset, steps=patches, verbose=1)
print(f"predictions: {predictions.shape}")

print("Writing predictions...")
writer = tf.io.TFRecordWriter(OUTPUT_IMAGE_FILE)

for i, prediction_patch in enumerate(predictions):
    if i == 0:
        print(f"Starting with patch {i}...")
        print(f"predictionPatch: {prediction_patch.shape}")

    if i % 50 == 0:
        print(f"Writing patch {i}...")

    prediction_patch = prediction_patch[
        x_buffer: x_buffer+Config.PATCH_SHAPE[0],
        y_buffer: y_buffer+Config.PATCH_SHAPE[1]
    ]

    example = tf.train.Example(
        features=tf.train.Features(
            feature={
            "prediction": tf.train.Feature(
                int64_list=tf.train.Int64List(
                    value=np.argmax(prediction_patch, axis=-1).flatten())),
            "cropland_etc": tf.train.Feature(
                float_list=tf.train.FloatList(
                    value=prediction_patch[:, :, 0:1].flatten())),
            "rice": tf.train.Feature(
                float_list=tf.train.FloatList(
                    value=prediction_patch[:, :, 1:2].flatten())),
            "forest": tf.train.Feature(
                float_list=tf.train.FloatList(
                    value=prediction_patch[:, :, 2:3].flatten())),
            "urban": tf.train.Feature(
                float_list=tf.train.FloatList(
                    value=prediction_patch[:, :, 3:4].flatten())),
            "others_etc": tf.train.Feature(
                float_list=tf.train.FloatList(
                    value=prediction_patch[:, :, 4:5].flatten())),
            }
        )
    )

    i += 1

    # Write the example.
    writer.write(example.SerializeToString())

writer.close()

# upload to gcp
upload_to_gcp = f"sudo gsutil cp {OUTPUT_IMAGE_FILE} {OUTPUT_GCS_PATH}"
result = subprocess.check_output(upload_to_gcp, shell=True)
print(f"uploading classified image to earth engine: {result}")

# upload to earth engine asset
upload_image = f"earthengine upload image --asset_id={Config.EE_OUTPUT_ASSET}/{Config.OUTPUT_NAME} --pyramiding_policy=mode {OUTPUT_GCS_PATH} {json_file}"
result = subprocess.check_output(upload_image, shell=True)
print(f"uploading classified image to earth engine: {result}")



# -*- coding: utf-8 -*-

import tensorflow as tf
from tensorflow import keras


class RemoteSensingFeatures:
    """
    A class for generating remote sensing features using TensorFlow.

    This class provides static methods to compute various remote sensing indices and concatenate them into feature tensors.
    """
    @staticmethod
    def normalized_difference(c1: tf.Tensor, c2: tf.Tensor, name: str = "nd") -> tf.Tensor:
        """
        Compute the normalized difference index between two spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the normalized difference index.

        """
        nd_f = keras.layers.Lambda(lambda x: ((x[0] - x[1]) / (x[0] + x[1])), name=name)([c1, c2])
        nd_inf = keras.layers.Lambda(lambda x: (x[0] - x[1]), name=f"{name}_inf")([c1, c2])
        return tf.where(tf.math.is_finite(nd_f), nd_f, nd_inf, name=name)

    @staticmethod
    def evi(c1: tf.Tensor, c2: tf.Tensor, c3: tf.Tensor, name: str = "evi") -> tf.Tensor:
        """
        Compute the enhanced vegetation index (EVI) using three spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            c3: A TensorFlow tensor representing the third spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the EVI.

        """
        _evi = keras.layers.Lambda(lambda x: 2.5 * ((x[0] - x[1]) / (x[0] + 6 * x[1] - 7.5 * x[2] + 1)), name=name)([c1, c2, c3])
        return _evi

    @staticmethod
    def savi(c1: tf.Tensor, c2: tf.Tensor, name: str = "savi") -> tf.Tensor:
        """
        Compute the soil-adjusted vegetation index (SAVI) between two spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the SAVI.

        """
        savi_f = keras.layers.Lambda(lambda x: ((x[0] - x[1]) / (x[0] + x[1] + 0.5)) * 1.5, name=name)([c1, c2])
        return savi_f

    @staticmethod
    def msavi(c1: tf.Tensor, c2: tf.Tensor, name: str = "msavi") -> tf.Tensor:
        """
        Compute the modified soil-adjusted vegetation index (MSAVI) between two spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the MSAVI.

        """
        msavi_f = keras.layers.Lambda(lambda x: (((2 * x[0] + 1) - tf.sqrt(((2 * x[0] + 1) * (2 * x[0] + 1)) - 8 * (x[0] - x[1]))) / 2), name=name)([c1, c2])
        return msavi_f

    @staticmethod
    def mtvi2(c1: tf.Tensor, c2: tf.Tensor, c3: tf.Tensor, name: str = "mtvi2") -> tf.Tensor:
        """
        Compute the modified transformed vegetation index 2 (MTVI2) using three spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            c3: A TensorFlow tensor representing the third spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the MTVI2.

        """
        mtvi2_f = keras.layers.Lambda(lambda x: (1.5 * (1.2 * (x[0] - x[2]) - 2.5 * (x[1] - x[2]))) / (tf.sqrt(((2 * x[0] + 1) * (2 * x[0] + 1)) - (6 * x[0] - 5 * tf.sqrt(x[1])) - 0.5)), name=name)([c1, c2, c3])
        return mtvi2_f

    @staticmethod
    def vari(c1: tf.Tensor, c2: tf.Tensor, c3: tf.Tensor, name: str = "vari") -> tf.Tensor:
        """
        Compute the visible atmospheric resistant index (VARI) using three spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            c3: A TensorFlow tensor representing the third spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the VARI.

        """
        vari_f = keras.layers.Lambda(lambda x: ((x[0] - x[1]) / (x[0] + x[1] - x[2])), name=name)([c1, c2, c3])
        vari_inf = keras.layers.Lambda(lambda x: (x[0] - x[1]), name=f"{name}_inf")([c1, c2, c3])
        return tf.where(tf.math.is_finite(vari_f), vari_f, vari_inf, name=name)

    @staticmethod
    def tgi(c1: tf.Tensor, c2: tf.Tensor, c3: tf.Tensor, name: str = "tgi") -> tf.Tensor:
        """
        Compute the triangular greenness index (TGI) using three spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            c3: A TensorFlow tensor representing the third spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the TGI.

        """
        tgi_f = keras.layers.Lambda(lambda x: ((120 * (x[1] - x[2])) - (190 * (x[1] - x[0]))) / 2, name=name)([c1, c2, c3])
        return tgi_f

    @staticmethod
    def ratio(c1: tf.Tensor, c2: tf.Tensor, name: str = "ratio") -> tf.Tensor:
        """
        Compute the ratio between two spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the ratio between the spectral bands.

        """
        ratio_f = keras.layers.Lambda(lambda x: x[0] / x[1], name=name)([c1, c2])
        ratio_inf = keras.layers.Lambda(lambda x: x[0], name=f"{name}_inf")([c1, c2])
        return tf.where(tf.math.is_finite(ratio_f), ratio_f, ratio_inf, name=name)

    @staticmethod
    def nvi(c1: tf.Tensor, c2: tf.Tensor, name: str = "nvi") -> tf.Tensor:
        """
        Compute the normalized vegetation index (NVI) between two spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the NVI.

        """
        nvi_f = keras.layers.Lambda(lambda x: x[0] / (x[0] + x[1]), name=name)([c1, c2])
        nvi_inf = keras.layers.Lambda(lambda x: x[0], name=f"{name}_inf")([c1, c2])
        return tf.where(tf.math.is_finite(nvi_f), nvi_f, nvi_inf, name=name)

    @staticmethod
    def diff_band(c1: tf.Tensor, c2: tf.Tensor, name: str = "diff") -> tf.Tensor:
        """
        Compute the difference between two spectral bands.

        Args:
            c1: A TensorFlow tensor representing the first spectral band.
            c2: A TensorFlow tensor representing the second spectral band.
            name: A string specifying the name for the operation.

        Returns:
            A TensorFlow tensor representing the difference between the spectral bands.

        """
        diff = keras.layers.Lambda(lambda x: x[0] - x[1], name=name)([c1, c2])
        return diff

    @staticmethod
    def concatenate_features_for_cnn(input_tensor: tf.Tensor) -> tf.Tensor:
        """
        Concatenate remote sensing features for Convolutional Neural Network (CNN) input.

        Args:
            input_tensor: A TensorFlow tensor representing the input remote sensing data.

        Returns:
            A TensorFlow tensor representing the concatenated features for CNN input.

        """
        red_before = input_tensor[:, :, :, 0:1]
        green_before = input_tensor[:, :, :, 1:2]
        blue_before = input_tensor[:, :, :, 2:3]
        nir_before = input_tensor[:, :, :, 3:4]
        red_during = input_tensor[:, :, :, 4:5]
        green_during = input_tensor[:, :, :, 5:6]
        blue_during = input_tensor[:, :, :, 6:7]
        nir_during = input_tensor[:, :, :, 7:8]

        ndvi_before = RemoteSensingFeatures.normalized_difference(nir_before, red_before, name="ndvi_before")
        ndvi_during = RemoteSensingFeatures.normalized_difference(nir_during, red_during, name="ndvi_during")
        evi_before = RemoteSensingFeatures.evi(nir_before, red_before, blue_before, name="evi_before")
        evi_during = RemoteSensingFeatures.evi(nir_during, red_during, blue_during, name="evi_during")
        ndwi_before = RemoteSensingFeatures.normalized_difference(green_before, nir_before, name="ndwi_before")
        ndwi_during = RemoteSensingFeatures.normalized_difference(green_during, nir_during, name="ndwi_during")
        savi_before = RemoteSensingFeatures.savi(nir_before, red_before, name="savi_before")
        savi_during = RemoteSensingFeatures.savi(nir_during, red_during, name="savi_during")
        # msavi_before = RemoteSensingFeatures.msavi(nir_before, red_before, name="msavi_before")
        # msavi_during = RemoteSensingFeatures.msavi(nir_during, red_during, name="msavi_during")
        mtvi2_before = RemoteSensingFeatures.mtvi2(nir_before, red_before, green_before, name="mtvi2_before")
        mtvi2_during = RemoteSensingFeatures.mtvi2(nir_during, red_during, green_during, name="mtvi2_during")
        # vari is not used because the computation gave some nan values
        # vari_before = RemoteSensingFeatures.vari(green_before, red_before, blue_before, name="vari_before")
        # vari_during = RemoteSensingFeatures.vari(green_during, red_during, blue_during, name="vari_during")
        # tgi_before = RemoteSensingFeatures.tgi(green_before, red_before, blue_before, name="tgi_before")
        # tgi_during = RemoteSensingFeatures.tgi(green_during, red_during, blue_during, name="tgi_during")

        red_diff1 = RemoteSensingFeatures.diff_band(red_before, red_during, name="diff1")
        green_diff1 = RemoteSensingFeatures.diff_band(green_before, green_during, name="diff2")
        blue_diff1 = RemoteSensingFeatures.diff_band(blue_before, blue_during, name="diff3")
        nir_diff1 = RemoteSensingFeatures.diff_band(nir_before, nir_during, name="diff4")

        return keras.layers.concatenate(
            [input_tensor, ndvi_before, ndvi_during, evi_before, evi_during, ndwi_before, ndwi_during, savi_before, savi_during,
             mtvi2_before, mtvi2_during,
             red_diff1, green_diff1, blue_diff1, nir_diff1],
            name="input_features"
        )

    @staticmethod
    def concatenate_features_for_dnn(input_tensor: tf.Tensor) -> tf.Tensor:
        """
        Concatenate remote sensing features for Deep Neural Network (DNN) input.

        Args:
            input_tensor: A TensorFlow tensor representing the input remote sensing data.

        Returns:
            A TensorFlow tensor representing the concatenated features for DNN input.

        """
        red_before = input_tensor[:, :, :, 0:1]
        green_before = input_tensor[:, :, :, 1:2]
        blue_before = input_tensor[:, :, :, 2:3]
        nir_before = input_tensor[:, :, :, 3:4]
        red_during = input_tensor[:, :, :, 4:5]
        green_during = input_tensor[:, :, :, 5:6]
        blue_during = input_tensor[:, :, :, 6:7]
        nir_during = input_tensor[:, :, :, 7:8]

        ndvi_before = RemoteSensingFeatures.normalized_difference(nir_before, red_before, name="ndvi_before")
        ndvi_during = RemoteSensingFeatures.normalized_difference(nir_during, red_during, name="ndvi_during")
        evi_before = RemoteSensingFeatures.evi(nir_before, red_before, blue_before, name="evi_before")
        evi_during = RemoteSensingFeatures.evi(nir_during, red_during, blue_during, name="evi_during")
        ndwi_before = RemoteSensingFeatures.normalized_difference(green_before, nir_before, name="ndwi_before")
        ndwi_during = RemoteSensingFeatures.normalized_difference(green_during, nir_during, name="ndwi_during")
        savi_before = RemoteSensingFeatures.savi(nir_before, red_before, name="savi_before")
        savi_during = RemoteSensingFeatures.savi(nir_during, red_during, name="savi_during")
        msavi_before = RemoteSensingFeatures.msavi(nir_before, red_before, name="msavi_before")
        msavi_during = RemoteSensingFeatures.msavi(nir_during, red_during, name="msavi_during")
        mtvi2_before = RemoteSensingFeatures.mtvi2(nir_before, red_before, green_before, name="mtvi2_before")
        mtvi2_during = RemoteSensingFeatures.mtvi2(nir_during, red_during, green_during, name="mtvi2_during")
        vari_before = RemoteSensingFeatures.vari(green_before, red_before, blue_before, name="vari_before")
        vari_during = RemoteSensingFeatures.vari(green_during, red_during, blue_during, name="vari_during")
        tgi_before = RemoteSensingFeatures.tgi(green_before, red_before, blue_before, name="tgi_before")
        tgi_during = RemoteSensingFeatures.tgi(green_during, red_during, blue_during, name="tgi_during")

        return keras.layers.concatenate(
            [input_tensor, ndvi_before, ndvi_during, evi_before, evi_during, ndwi_before, ndwi_during, savi_before, savi_during,
             msavi_before, msavi_during, mtvi2_before, mtvi2_during, vari_before, vari_during, tgi_before, tgi_during],
            name="input_features"
        )

    @staticmethod
    def derive_features_for_dnn(features_dict: dict, added_features: list = []) -> dict:
        """
        Concatenate remote sensing features for Deep Neural Network (DNN) input.

        Args:
            input_tensor: A TensorFlow tensor representing the input remote sensing data.

        Returns:
            A TensorFlow tensor representing the concatenated features for DNN input.

        """
        red_before = features_dict["red_before"]
        green_before = features_dict["green_before"]
        blue_before = features_dict["blue_before"]
        nir_before = features_dict["nir_before"]
        red_during = features_dict["red_during"]
        green_during = features_dict["green_during"]
        blue_during = features_dict["blue_during"]
        nir_during = features_dict["nir_during"]

        feature_col = {
            "ndvi_before": RemoteSensingFeatures.normalized_difference(nir_before, red_before, name="ndvi_before"),
            "ndvi_during": RemoteSensingFeatures.normalized_difference(nir_during, red_during, name="ndvi_during"),
            "evi_before": RemoteSensingFeatures.evi(nir_before, red_before, blue_before, name="evi_before"),
            "evi_during": RemoteSensingFeatures.evi(nir_during, red_during, blue_during, name="evi_during"),
            "ndwi_before": RemoteSensingFeatures.normalized_difference(green_before, nir_before, name="ndwi_before"),
            "ndwi_during": RemoteSensingFeatures.normalized_difference(green_during, nir_during, name="ndwi_during"),
            "savi_before": RemoteSensingFeatures.savi(nir_before, red_before, name="savi_before"),
            "savi_during": RemoteSensingFeatures.savi(nir_during, red_during, name="savi_during"),
            "msavi_before": RemoteSensingFeatures.msavi(nir_before, red_before, name="msavi_before"),
            "msavi_during": RemoteSensingFeatures.msavi(nir_during, red_during, name="msavi_during"),
            "mtvi2_before": RemoteSensingFeatures.mtvi2(nir_before, red_before, green_before, name="mtvi2_before"),
            "mtvi2_during": RemoteSensingFeatures.mtvi2(nir_during, red_during, green_during, name="mtvi2_during"),
            "vari_before": RemoteSensingFeatures.vari(green_before, red_before, blue_before, name="vari_before"),
            "vari_during": RemoteSensingFeatures.vari(green_during, red_during, blue_during, name="vari_during"),
            "tgi_before": RemoteSensingFeatures.tgi(green_before, red_before, blue_before, name="tgi_before"),
            "tgi_during": RemoteSensingFeatures.tgi(green_during, red_during, blue_during, name="tgi_during"),
        }

        for feature in added_features:
            features_dict[feature] = feature_col[feature]

        return features_dict



try:
    from aces.ee_utils import EEUtils
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

    from aces.ee_utils import EEUtils

import ee

EEUtils.initialize_session(use_highvolume=True)

country = "Bhutan"
region = "Paro"
year = 2021

dem = ee.Image("MERIT/DEM/v1_0_3")

# baseModule = require("users/biplovbhandari/ACES_2:main.js")

planet_asia = ee.ImageCollection("projects/planet-nicfi/assets/basemaps/asia")

fc_country = ee.FeatureCollection(EEUtils.country_bbox(country))

admin_1 = ee.FeatureCollection("projects/servir-sco-assets/assets/Bhutan/BT_Admin_1")
paro = admin_1.filter(ee.Filter.eq("ADM1_EN", "Paro"))
punakha = admin_1.filter(ee.Filter.eq("ADM1_EN", "Punakha"))
punakha = punakha.map(lambda f: f.set("reduceId", 1))

monggar = admin_1.filter(ee.Filter.eq("ADM1_EN", "Monggar"))
monggar = monggar.map(lambda f: f.set("reduceId", 1))

name_region = {
    "Punakha": punakha,
    "Paro": paro,
    "Monggar": monggar
}

region_fc = name_region[region]

# Rice growing zone
dem_punakha = dem.clip(punakha)
rice_zone1_punakha = dem_punakha.gt(1500).And(dem_punakha.lte(2600))
rice_zone2_punakha = dem_punakha.gt(1000).And(dem_punakha.lte(1500))
rice_zone_punakha = dem_punakha.gt(1000).And(dem_punakha.lte(2600))
rice_zone_punakha = rice_zone_punakha.selfMask()

punakha_image = punakha.reduceToImage(properties = ["reduceId"], reducer = ee.Reducer.first())

planet_asia = planet_asia.map(lambda img: img.multiply(0.0001).copyProperties(img, img.toDictionary().keys())
                         .set("system:time_start", img.get("system:time_start"), "system:time_end", img.get("system:time_end"))
                        )


# Export Planet Mosaic

# before growing season
# mar, apr, may
img_mar = planet_asia.filterDate(ee.Date.fromYMD(year, 3, 1), ee.Date.fromYMD(year, 3, 1).advance(1, "month").advance(-1, "day")).first()
img_apr = planet_asia.filterDate(ee.Date.fromYMD(year, 4, 1), ee.Date.fromYMD(year, 4, 1).advance(1, "month").advance(-1, "day")).first()
img_may = planet_asia.filterDate(ee.Date.fromYMD(year, 5, 1), ee.Date.fromYMD(year, 5, 1).advance(1, "month").advance(-1, "day")).first()
img_jun = planet_asia.filterDate(ee.Date.fromYMD(year, 6, 1), ee.Date.fromYMD(year, 6, 1).advance(1, "month").advance(-1, "day")).first()

before_collection = ee.ImageCollection([img_mar, img_apr, img_may])
print("collection_before", before_collection.size().getInfo())
before_composite = before_collection.median()

before_reducers = EEUtils.get_image_collection_statistics(before_collection)

before_indices = before_collection.map(EEUtils.calculate_planet_indices)
before_indices_reducers = EEUtils.get_image_collection_statistics(before_indices)
before_indices = before_indices.median()

before_composite = before_composite.select(["R", "G", "B", "N"], ["red", "green", "blue", "nir"]) \
                          .addBands(before_reducers).addBands(before_indices).addBands(before_indices_reducers)

print("before_composite", before_composite.bandNames().getInfo())


# during gwoing season
# july, aug, sept
img_jul = planet_asia.filterDate(ee.Date.fromYMD(year, 7, 1), ee.Date.fromYMD(year, 7, 1).advance(1, "month").advance(-1, "day")).first()
img_aug = planet_asia.filterDate(ee.Date.fromYMD(year, 8, 1), ee.Date.fromYMD(year, 8, 1).advance(1, "month").advance(-1, "day")).first()
img_sep = planet_asia.filterDate(ee.Date.fromYMD(year, 9, 1), ee.Date.fromYMD(year, 9, 1).advance(1, "month").advance(-1, "day")).first()

during_collection = ee.ImageCollection([img_jul, img_aug, img_sep])
print("during_collection", during_collection.size().getInfo())
during_composite = during_collection.median()

during_reducers = EEUtils.get_image_collection_statistics(during_collection)

during_indices = during_collection.map(EEUtils.calculate_planet_indices)
during_indices_reducers = EEUtils.get_image_collection_statistics(during_indices)
during_indices = during_indices.median()

during_composite = during_composite.select(["R", "G", "B", "N"], ["red", "green", "blue", "nir"]) \
    .addBands(during_reducers).addBands(during_indices).addBands(during_indices_reducers)

print("during_composite", during_composite.bandNames().getInfo())


# after growing season
# oct, nov, dec
img_oct = planet_asia.filterDate(ee.Date.fromYMD(year, 10, 1), ee.Date.fromYMD(year, 10, 1).advance(1, "month").advance(-1, "day")).first()
img_nov = planet_asia.filterDate(ee.Date.fromYMD(year, 11, 1), ee.Date.fromYMD(year, 11, 1).advance(1, "month").advance(-1, "day")).first()
img_dec = planet_asia.filterDate(ee.Date.fromYMD(year, 12, 1), ee.Date.fromYMD(year, 12, 1).advance(1, "month").advance(-1, "day")).first()

after_collection = ee.ImageCollection([img_oct, img_nov, img_dec])
print("after_collection", after_collection.size().getInfo())
after_composite = after_collection.median()

after_reducers = EEUtils.get_image_collection_statistics(after_collection)

after_indices = after_collection.map(EEUtils.calculate_planet_indices)
after_indices_reducers = EEUtils.get_image_collection_statistics(after_indices)
after_indices = after_indices.median()

after_composite = after_composite.select(["R", "G", "B", "N"], ["red", "green", "blue", "nir"]) \
    .addBands(after_reducers).addBands(after_indices).addBands(after_indices_reducers)

print("after_composite", after_composite.bandNames().getInfo())
print("region_fc.geometry().bounds().getInfo()['coordinates']", region_fc.geometry().bounds().getInfo()["coordinates"])
export_kwargs = {
    "region": region_fc.geometry().bounds().getInfo()["coordinates"],
    "scale": 5,
}

EEUtils.export_image(before_composite, export_type="asset",
                     **{**export_kwargs, "asset_id": f"projects/servir-sco-assets/assets/Bhutan/ACES_2/{region}_Composite_{year}/composite_before",
                     "description": f"{year}_composite_before"}
                    )
EEUtils.export_image(during_composite, export_type="asset",
                     **{**export_kwargs, "asset_id": f"projects/servir-sco-assets/assets/Bhutan/ACES_2/{region}_Composite_{year}/composite_during",
                        "description": f"{year}_composite_during"}
                    )
EEUtils.export_image(after_composite, export_type="asset",
                     **{**export_kwargs, "asset_id": f"projects/servir-sco-assets/assets/Bhutan/ACES_2/{region}_Composite_{year}/composite_after",
                        "description": f"{year}_composite_after"}
                    )



# -*- coding: utf-8 -*-

"""
model_builder.py:

Model Builder Class for Creating and Compiling Neural Network Models

This module provides a `ModelBuilder` class that is responsible for creating and compiling neural network models.
It includes methods for building and compiling models of different types, such as Deep Neural Network (**DNN**),
Convolutional Neural Network (**CNN**), and **U-Net**, based on the
provided specifications. The class also contains utility methods for constructing custom layers and defining metrics.
"""

import logging
logging.basicConfig(level=logging.INFO)

import tensorflow as tf
from tensorflow import keras

from aces.metrics import Metrics
from aces.remote_sensing import RemoteSensingFeatures as rs


class ModelBuilder:
    """
    ModelBuilder Class for Creating and Compiling Neural Network Models

    This class provides methods for building and compiling neural network models of different types, such as DNN, CNN, and U-Net,
    based on the provided specifications. It includes utility methods for constructing custom layers and defining metrics.

    Attributes:
        in_size (int): The input size of the models.
        out_classes (int): The number of output classes for classification models.
        optimizer (tf.keras.optimizers.Optimizer): The optimizer to use for model compilation.
        loss (tf.keras.losses.Loss): The loss function to use for model compilation.

    Methods:
        build_model(model_type, **kwargs):
            Builds and compiles a neural network model based on the provided model type.
        build_and_compile_dnn_model(**kwargs):
            Builds and compiles a Deep Neural Network (DNN) model.
        build_and_compile_cnn_model(**kwargs):
            Builds and compiles a Convolutional Neural Network (CNN) model.
        build_and_compile_unet_model(**kwargs):
            Builds and compiles a U-Net model.
        _build_and_compile_unet_model(**kwargs):
            Helper method for building and compiling a U-Net model.
    """

    def __init__(self, features, out_classes, optimizer, loss):
        """
        Initialize ModelBuilder with input size, output classes, optimizer, and loss.

        Args:
            in_size (int): The input size of the models.
            out_classes (int): The number of output classes for classification models.
            optimizer (tf.keras.optimizers.Optimizer): The optimizer to use for model compilation.
            loss (tf.keras.losses.Loss): The loss function to use for model compilation.
        """
        self.features = features
        self.in_size = len(features)
        self.out_classes = out_classes
        self.optimizer = optimizer
        self.loss = loss

    def build_model(self, model_type, **kwargs):
        """
        Builds and compiles a neural network model based on the provided model type.

        Args:
            model_type (str): The type of the model to build ("dnn", "cnn", "unet").
            **kwargs: Additional keyword arguments specific to the model type.

        Returns:
            keras.Model: The compiled neural network model.

        Raises:
            ValueError: If an invalid model type is provided.
        """
        FOR_AI_PLATFORM = kwargs.get("FOR_AI_PLATFORM", False)
        if model_type == "dnn":
            if FOR_AI_PLATFORM:
                return self.build_and_compile_dnn_model_for_ai_platform(**kwargs)
            else:
                return self.build_and_compile_dnn_model(**kwargs)
        elif model_type == "cnn":
            return self.build_and_compile_cnn_model(**kwargs)
        elif model_type == "unet":
            if FOR_AI_PLATFORM:
                return self.build_and_compile_unet_model_for_ai_platform(**kwargs)
            else:
                return self.build_and_compile_unet_model(**kwargs)
        else:
            raise ValueError(f"Invalid model type: {model_type}")

    def build_and_compile_dnn_model(self, **kwargs):
        """
        Builds and compiles a Deep Neural Network (DNN) model.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            keras.Model: The compiled DNN model.
        """
        INITIAL_BIAS = kwargs.get("INITIAL_BIAS", None)
        logging.info(f"INITIAL_BIAS: {INITIAL_BIAS}")

        if INITIAL_BIAS is not None:
            INITIAL_BIAS = tf.keras.initializers.Constant(INITIAL_BIAS)
        else:
            INITIAL_BIAS = "zeros"

        inputs = keras.Input(shape=(None, self.in_size), name="input_layer")

        x = keras.layers.Dense(256, activation="relu")(inputs)
        x = keras.layers.Dropout(0.2)(x)
        x = keras.layers.Dense(128, activation="relu")(x)
        x = keras.layers.Dropout(0.2)(x)
        x = keras.layers.Dense(64, activation="relu")(x)
        x = keras.layers.Dropout(0.2)(x)
        x = keras.layers.Dense(32, activation="relu")(x)
        x = keras.layers.Dropout(0.2)(x)
        output = keras.layers.Dense(self.out_classes, activation=kwargs.get("ACTIVATION_FN"), bias_initializer=INITIAL_BIAS)(x)

        model = keras.models.Model(inputs=inputs, outputs=output)
        metrics_list = [
            Metrics.precision(),
            Metrics.recall(),
            keras.metrics.CategoricalAccuracy(),
            Metrics.one_hot_io_u(self.out_classes),
        ]

        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=metrics_list)
        return model


    def build_and_compile_dnn_model_for_ai_platform(self, **kwargs):
        """
        Builds and compiles a Deep Neural Network (DNN) model.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            keras.Model: The compiled DNN model.
        """
        INITIAL_BIAS = kwargs.get("INITIAL_BIAS", None)
        logging.info(f"INITIAL_BIAS: {INITIAL_BIAS}")
        # DNN_DURING_ONLY = kwargs.get("DURING_ONLY", False)

        DERIVE_FEATURES = kwargs.get("DERIVE_FEATURES", False)

        if INITIAL_BIAS is not None:
            INITIAL_BIAS = tf.keras.initializers.Constant(INITIAL_BIAS)
        else:
            INITIAL_BIAS = "zeros"

        inputs_main = keras.Input(shape=(None, None, self.in_size), name="input_layer")

        if DERIVE_FEATURES:
            inputs = rs.concatenate_features_for_dnn(inputs_main)
        else:
            inputs = inputs_main

        x = keras.layers.Conv2D(256, (1, 1), activation="relu")(inputs)
        x = keras.layers.Dropout(0.2)(x)
        x = keras.layers.Conv2D(128, (1, 1), activation="relu")(x)
        x = keras.layers.Dropout(0.2)(x)
        x = keras.layers.Conv2D(64, (1, 1), activation="relu")(x)
        x = keras.layers.Dropout(0.2)(x)
        x = keras.layers.Conv2D(32, (1, 1), activation="relu")(x)
        x = keras.layers.Dropout(0.2)(x)
        output = keras.layers.Conv2D(self.out_classes, (1, 1), activation=kwargs.get("ACTIVATION_FN"), bias_initializer=INITIAL_BIAS)(x)

        model = keras.models.Model(inputs=inputs_main, outputs=output)

        wrapped_model = ModelWrapper(ModelPreprocess(self.features), model)

        metrics_list = [
            Metrics.precision(),
            Metrics.recall(),
            keras.metrics.CategoricalAccuracy(),
            Metrics.one_hot_io_u(self.out_classes),
        ]
        wrapped_model.compile(optimizer=self.optimizer, loss=self.loss, metrics=metrics_list)
        return model, wrapped_model

    def build_and_compile_cnn_model(self, **kwargs):
        """
        Builds and compiles a Convolutional Neural Network (CNN) model.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            keras.Model: The compiled CNN model.
        """
        inputs = keras.Input(shape=(kwargs.get("PATCH_SHAPE", 128)[0], kwargs.get("PATCH_SHAPE", 128)[0], self.in_size))
        x = keras.layers.Conv2D(32, 3, activation="relu", name="convd-1", padding="same")(inputs)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Dropout(0.15)(x)
        x = keras.layers.Conv2D(64, 3, activation="relu", name="convd-2", padding="same")(x)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Dropout(0.15)(x)
        x = keras.layers.Conv2D(128, 3, activation="relu", name="convd-3", padding="same")(x)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Dropout(0.15)(x)
        x = keras.layers.Conv2D(128, 3, activation="relu", name="convd-4", padding="same")(x)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Dropout(0.15)(x)
        outputs = keras.layers.Conv2D(self.out_classes, (1, 1), activation="softmax", name="final_conv")(x)
        model = keras.Model(inputs, outputs, name="cnn_model")

        metrics_list = [
            Metrics.precision(),
            Metrics.recall(),
            keras.metrics.CategoricalAccuracy(),
            Metrics.one_hot_io_u(self.out_classes),
        ]

        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=metrics_list)
        return model

    def build_and_compile_unet_model_for_ai_platform(self, **kwargs):
        """
        Builds and compiles a U-Net model.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            keras.Model: The compiled U-Net model.
        """
        if len(kwargs.get("physical_devices")) > 0:
            with tf.distribute.MirroredStrategy().scope():
                return self._build_and_compile_unet_model_for_ai_plaform(**kwargs)
        else:
            logging.info("No distributed strategy found.")
            return self._build_and_compile_unet_model(**kwargs)

    def _build_and_compile_unet_model_for_ai_plaform(self, **kwargs):
        """
        Helper method for building and compiling a U-Net model.
        """
        model = self._build_and_compile_unet_model(**kwargs)
        wrapped_model = ModelWrapper(ModelPreprocess(self.features), model)
        metrics_list = [
            Metrics.precision(),
            Metrics.recall(),
            keras.metrics.CategoricalAccuracy(),
            Metrics.one_hot_io_u(self.out_classes),
        ]
        wrapped_model.compile(optimizer=self.optimizer, loss=self.loss, metrics=metrics_list)
        return model, wrapped_model

    def build_and_compile_unet_model(self, **kwargs):
        """
        Builds and compiles a U-Net model.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            keras.Model: The compiled U-Net model.
        """
        if len(kwargs.get("physical_devices")) > 0:
            with tf.distribute.MirroredStrategy().scope():
                model = self._build_and_compile_unet_model(**kwargs)
                metrics_list = [
                    Metrics.precision(),
                    Metrics.recall(),
                    keras.metrics.CategoricalAccuracy(),
                    Metrics.one_hot_io_u(self.out_classes),
                ]
                model.compile(optimizer=self.optimizer, loss=self.loss, metrics=metrics_list)
                return model
        else:
            logging.info("No distributed strategy found.")
            model = self._build_and_compile_unet_model(**kwargs)
            metrics_list = [
                Metrics.precision(),
                Metrics.recall(),
                keras.metrics.CategoricalAccuracy(),
                Metrics.one_hot_io_u(self.out_classes),
            ]
            model.compile(optimizer=self.optimizer, loss=self.loss, metrics=metrics_list)
            return model

    def _build_and_compile_unet_model(self, **kwargs):
        """
        Helper method for building and compiling a U-Net model.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            keras.Model: The compiled U-Net model.
        """
        inputs = keras.Input(shape=(None, None, self.in_size))

        DERIVE_FEATURES = kwargs.get("DERIVE_FEATURES", False)

        print(f"DERIVE_FEATURES: {DERIVE_FEATURES}")

        if DERIVE_FEATURES:
            input_features = rs.concatenate_features_for_cnn(inputs)
        else:
            input_features = inputs


        x = keras.layers.Conv2D(32, 3, strides=2, padding="same")(input_features)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Activation("relu")(x)

        previous_block_activation = x

        l2_regularizer = keras.regularizers.l2(0.001)

        for filters in [64, 128, 256]:
            x = keras.layers.Activation("relu")(x)
            x = keras.layers.SeparableConv2D(filters, 3, padding="same")(x)
            x = keras.layers.BatchNormalization()(x)

            x = keras.layers.Activation("relu")(x)
            x = keras.layers.SeparableConv2D(filters, 3, padding="same")(x)
            x = keras.layers.BatchNormalization()(x)

            x = keras.layers.MaxPooling2D(3, strides=2, padding="same")(x)

            residual = keras.layers.Conv2D(filters, 1, strides=2, padding="same")(
                previous_block_activation
            )
            x = keras.layers.add([x, residual])
            previous_block_activation = x

        for filters in [256, 128, 64, 32]:
            x = keras.layers.Activation("relu")(x)
            x = keras.layers.Conv2DTranspose(filters, 3, padding="same")(x)
            x = keras.layers.BatchNormalization()(x)

            x = keras.layers.Activation("relu")(x)
            x = keras.layers.Conv2DTranspose(filters, 3, padding="same")(x)
            x = keras.layers.BatchNormalization()(x)

            x = keras.layers.UpSampling2D(2)(x)

            residual = keras.layers.UpSampling2D(2)(previous_block_activation)
            residual = keras.layers.Conv2D(filters, 1, padding="same")(residual)
            x = keras.layers.add([x, residual])
            previous_block_activation = x

        outputs = keras.layers.Conv2D(self.out_classes, 3, activation=kwargs.get("ACTIVATION_FN"), padding="same", name="final_conv")(x)

        model = keras.Model(inputs=inputs, outputs=outputs, name="unet")

        return model

    def _build_and_compile_vanilla_unet_model(self, **kwargs):
        """
        Helper method for building and compiling a U-Net model.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            keras.Model: The compiled U-Net model.
        """
        inputs = keras.Input(shape=(None, None, self.in_size))

        DERIVE_FEATURES = kwargs.get("DERIVE_FEATURES", False)

        print(f"DERIVE_FEATURES: {DERIVE_FEATURES}")

        if DERIVE_FEATURES:
            input_features = rs.concatenate_features_for_cnn(inputs)
        else:
            input_features = inputs

        c1 = keras.layers.Conv2D(16, (3, 3), padding="same")(input_features)
        c1 = keras.layers.BatchNormalization()(c1)
        c1 = keras.layers.Activation("relu")(c1)
        p1 = keras.layers.MaxPooling2D((2, 2))(c1)

        c2 = keras.layers.Conv2D(32, (3, 3), padding="same")(p1)
        c2 = keras.layers.BatchNormalization()(c2)
        c2 = keras.layers.Activation("relu")(c2)
        p2 = keras.layers.MaxPooling2D((2, 2))(c2)

        c3 = keras.layers.Conv2D(64, (3, 3), padding="same")(p2)
        c3 = keras.layers.BatchNormalization()(c3)
        c3 = keras.layers.Activation("relu")(c3)
        p3 = keras.layers.MaxPooling2D((2, 2))(c3)

        c4 = keras.layers.Conv2D(128, (3, 3), padding="same")(p3)
        c4 = keras.layers.BatchNormalization()(c4)
        c4 = keras.layers.Activation("relu")(c4)
        p4 = keras.layers.MaxPooling2D((2, 2))(c4)

        c5 = keras.layers.Conv2D(256, (3, 3), padding="same")(p4)
        c5 = keras.layers.BatchNormalization()(c5)
        c5 = keras.layers.Activation("relu")(c5)

        # Decoder
        u6 = keras.layers.UpSampling2D((2, 2))(c5)
        u6 = keras.layers.Conv2D(128, (3, 3), padding="same")(u6)
        u6 = keras.layers.BatchNormalization()(u6)
        u6 = keras.layers.Activation("relu")(u6)
        u6 = keras.layers.Add()([u6, c4])

        u7 = keras.layers.UpSampling2D((2, 2))(u6)
        u7 = keras.layers.Conv2D(64, (3, 3), padding="same")(u7)
        u7 = keras.layers.BatchNormalization()(u7)
        u7 = keras.layers.Activation("relu")(u7)
        u7 = keras.layers.Add()([u7, c3])

        u8 = keras.layers.UpSampling2D((2, 2))(u7)
        u8 = keras.layers.Conv2D(32, (3, 3), padding="same")(u8)
        u8 = keras.layers.BatchNormalization()(u8)
        u8 = keras.layers.Activation("relu")(u8)
        u8 = keras.layers.Add()([u8, c2])

        u9 = keras.layers.UpSampling2D((2, 2))(u8)
        u9 = keras.layers.Conv2D(16, (3, 3), padding="same")(u9)
        u9 = keras.layers.BatchNormalization()(u9)
        u9 = keras.layers.Activation("relu")(u9)
        u9 = keras.layers.Add()([u9, c1])

        # Output Layer
        output_layer = keras.layers.Conv2D(self.out_classes, 3, activation=kwargs.get("ACTIVATION_FN"), padding="same", name="final_conv")(u9)

        # Build Model
        model = keras.Model(inputs=inputs, outputs=output_layer, name="vanilla_unet")

        return model



class AddExtraFeatures(tf.keras.layers.Layer):
    def __init__(self, added_features):
        super().__init__()
        self.added_features = added_features

    def call(self, features_dict, labels):
        features_dict = rs.derive_features_for_dnn(features_dict, self.added_features)
        return features_dict, labels



# A Layer to stack and reshape the input tensors.
class ModelPreprocess(keras.layers.Layer):
    def __init__(self, in_features, **kwargs):
        self.in_features = in_features
        super(ModelPreprocess, self).__init__(**kwargs)

    def call(self, features_dict):
        # (None, 1, 1, 1) -> (None, 1, 1, P) for dnn
        # (None, H, W, 1) -> (None, H, W, P) for cnn/unet
        return tf.concat([features_dict[b] for b in self.in_features], axis=3)

    def get_config(self):
        config = super().get_config()
        return config


# A Model that wraps the base model with the preprocessing layer.
class ModelWrapper(keras.Model):
    def __init__(self, preprocessing, backbone, **kwargs):
        super().__init__(**kwargs)
        self.preprocessing = preprocessing
        self.backbone = backbone

    def call(self, features_dict):
        x = self.preprocessing(features_dict)
        return self.backbone(x)

    def get_config(self):
        config = super().get_config()
        return config

@tf.autograph.experimental.do_not_convert
class DeSerializeInput(tf.keras.layers.Layer):
    def __init__(self, in_features, **kwargs):
        self.in_features = in_features
        super().__init__(**kwargs)

    def call(self, inputs_dict):
        deserialize = {
            k: tf.map_fn(lambda x: tf.io.parse_tensor(x, tf.float32),
                         tf.io.decode_base64(v),
                         fn_output_signature=tf.float32)
            for (k, v) in inputs_dict.items() if k in self.in_features
        }
        return deserialize

    def get_config(self):
        config = super().get_config()
        return config


@tf.autograph.experimental.do_not_convert
class ReSerializeOutput(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def call(self, output_tensor, name):
        if not name: name = "output"
        return {name: tf.identity(tf.map_fn(
            lambda x: tf.io.encode_base64(tf.io.serialize_tensor(x)),
            output_tensor, fn_output_signature=tf.string
            ),
            name=name
        )}

    def get_config(self):
        config = super().get_config()
        return config



# -*- coding: utf-8 -*-

try:
    from aces.ee_utils import EEUtils
    from aces.config import Config
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

    from aces.ee_utils import EEUtils
    from aces.config import Config

import ee

EEUtils.initialize_session(use_highvolume=True)

country = "Bhutan"
region = "Paro"
year = 2021
sensor = "planet"

fc_country = ee.FeatureCollection(EEUtils.country_bbox(country))
admin_1 = ee.FeatureCollection("projects/servir-sco-assets/assets/Bhutan/BT_Admin_1")
paro = admin_1.filter(ee.Filter.eq("ADM1_EN", "Paro"))
punakha = admin_1.filter(ee.Filter.eq("ADM1_EN", "Punakha"))
punakha = punakha.map(lambda f: f.set("reduceId", 1))

monggar = admin_1.filter(ee.Filter.eq("ADM1_EN", "Monggar"))
monggar = monggar.map(lambda f: f.set("reduceId", 1))

name_region = {
    "Punakha": punakha,
    "Paro": paro,
    "Monggar": monggar
}

rice_zone = {
    "Punakha": {
        "min": 1000,
        "max": 2600,
    },
    "Paro": {
        "min": 1500,
        "max": 2600,
    },
}

region_fc = name_region[region]

# smaller example region
region_fc = ee.Geometry.Polygon(
        [[[89.25910506388209, 27.58540960195346],
          [89.25910506388209, 27.159794800895543],
          [89.58182845255396, 27.159794800895543],
          [89.58182845255396, 27.58540960195346]]], None, False);

composite_before_paro = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/Paro_Rice_Composite_2021/composite_before")
composite_during_paro = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/Paro_Rice_Composite_2021/composite_during")
composite_before_punakha = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/Punakha_Composite_2021/composite_before")
composite_during_punakha = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/Punakha_Composite_2021/composite_during")

name_image = {
    "Punakha": [composite_before_punakha, composite_during_punakha],
    "Paro": [composite_before_paro, composite_during_paro],
}

composite_before = name_image[region][0]
composite_during = name_image[region][1]

if sensor == "planet":
    bands = ["red", "green", "blue", "nir"]

composite_before = composite_before.select(bands)
composite_during = composite_during.select(bands)

composite_before = composite_before.regexpRename("$(.*)", "_before")
composite_during = composite_during.regexpRename("$(.*)", "_during")
image = composite_before.addBands(composite_during).toFloat()

if Config.USE_ELEVATION:
    elevation = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/elevationParo")
    slope = ee.Image("projects/servir-sco-assets/assets/Bhutan/ACES_2/slopeParo")
    image = image.addBands(elevation).addBands(slope).toFloat()
    Config.FEATURES.extend(["elevation", "slope"])


if Config.USE_S1:
    sentinel1_asc_before_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Ascending2021/s1AscBefore")
    sentinel1_asc_during_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Ascending2021/s1AscDuring")
    sentinel1_desc_before_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Descending2021/s1DescBefore")
    sentinel1_desc_during_composite = ee.Image("projects/servir-sco-assets/assets/Bhutan/Sentinel1Descending2021/s1DescDuring")

    image = image.addBands(sentinel1_asc_before_composite).addBands(sentinel1_asc_during_composite).addBands(sentinel1_desc_before_composite).addBands(sentinel1_desc_during_composite).toFloat()
    Config.FEATURES.extend(["vv_asc_before", "vh_asc_before", "vv_asc_during", "vh_asc_during",
                            "vv_desc_before", "vh_desc_before", "vv_desc_during", "vh_desc_during"])

# dem = ee.Image("MERIT/DEM/v1_0_3") # ee.Image('USGS/SRTMGL1_003');
# dem = dem.clip(fc_country)
# riceZone = dem.gt(rice_zone[region]["min"]).And(dem.lte(rice_zone[region]["max"]))
# image = image.clip(region_fc).updateMask(riceZone)

image = image.select(Config.FEATURES)
print("image", image.bandNames().getInfo())

# Specify patch and file dimensions.
formatOptions = {
  "patchDimensions": [Config.PATCH_SHAPE_SINGLE, Config.PATCH_SHAPE_SINGLE],
  "maxFileSize": 104857600,
  "compressed": True
}

if Config.KERNEL_BUFFER:
    formatOptions["kernelSize"] = Config.KERNEL_BUFFER

# Setup the task
image_export_options = {
    "description": Config.GCS_IMAGE_DIR.split("/")[-1],
    "file_name_prefix": f"{Config.GCS_IMAGE_DIR}/{Config.GCS_IMAGE_PREFIX}",
    "bucket": Config.GCS_BUCKET,
    "scale": Config.SCALE,
    "file_format": "TFRecord",
    "region": region_fc, # image.geometry(),
    "format_options": formatOptions,
    "max_pixels": 1e13,
}

print("image_export_options", image_export_options)

EEUtils.export_image(image, export_type=["cloud"], start_training=True, **image_export_options)



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)

try:
    from aces.config import Config
    from aces.data_processor import DataProcessor
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

    from aces.config import Config
    from aces.data_processor import DataProcessor



if __name__ == "__main__":
    config_file = "config.env"
    config = Config(config_file)
    # expand the config
    additional_config = {
        "PRINT_DATASET": True
    }
    n_training_records, n_testing_records, n_validation_records = DataProcessor.calculate_n_samples(**{**config.__dict__, **additional_config})
    logging.info(f"no of training records: {n_training_records}")
    logging.info(f"no of testing records: {n_testing_records}")
    logging.info(f"no of validation records: {n_validation_records}")



# -*- coding: utf-8 -*-

"""
metrics.py: Custom Metrics for Model Evaluation and Utility Functions for Model Visualization

This module provides a collection of custom metrics that can be used for evaluating model performance in tasks such as
image segmentation. This module includes a wide variety of evaluation metrics listed below.
Additionally, it contains utility functions for plotting and visualizing model metrics during training.
"""

import logging
logging.basicConfig(level=logging.INFO)

import matplotlib.pyplot as plt

from tensorflow import keras
from keras import backend as K

__all__ = ["Metrics"]

class Metrics:
    """
    A class containing various metrics functions for model evaluation.

    This class provides a collection of static methods, each representing a different evaluation metric. These metrics can be used to
    evaluate the performance of classification or segmentation models during training or testing.

    Methods:
        precision():
            Computes the precision metric.
        recall():
            Computes the recall metric.
        f1_m():
            Computes the F1 score metric. [van1979information, chicco2020advantages]
        one_hot_io_u(num_classes):
            Computes the Intersection over Union (IoU) metric for each class. [jaccard1901etude]
        true_positives():
            Computes the true positives metric also known as sensitivity.[yerushalmy1947statistical]
        false_positives():
            Computes the false positives metric. [cohen2013statistical]
        true_negatives():
            Computes the true negatives metric also known as specificity. [yerushalmy1947statistical]
        false_negatives():
            Computes the false negatives metric. [cohen2013statistical]
        binary_accuracy():
            Computes the binary accuracy metric.
        auc():
            Computes the Area Under the Curve (AUC) metric. [provost2001robust]
        prc():
            Computes the Precision-Recall Curve (PRC) metric.
        dice_coef():
            Computes the Dice coefficient metric. [milletari2016v]
        dice_loss():
            Computes the Dice loss metric. [sudre2017generalised]
        bce_loss():
            Computes the Binary Cross-Entropy (BCE) loss metric. [zhu2018negative; yi2004automated]
        bce_dice_loss():
            Computes the BCE and Dice loss metric. cite [taghanaki2019combo]
        tversky():
            Computes the Tversky index metric.[tversky1977features]
        tversky_loss():
            Computes the Tversky loss metric. [salehi2017tversky]
        focal_tversky_loss():
            Computes the focal Tversky loss metric [abraham2019novel]
    """

    @staticmethod
    def recall_m(y_true, y_pred):
        """
        Calculate the recall metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.

        Returns:
            Recall metric value.
        """
        true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
        possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
        recall = true_positives / (possible_positives + K.epsilon())
        return recall

    @staticmethod
    def precision_m(y_true, y_pred):
        """
        Calculate the precision metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.

        Returns:
            Precision metric value.
        """
        true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
        predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
        precision = true_positives / (predicted_positives + K.epsilon())
        return precision

    @staticmethod
    def f1_m(y_true, y_pred):
        """
        Calculate the F1-score metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.

        Returns:
            F1-score metric value.
        """
        precision = Metrics.precision_m(y_true, y_pred)
        recall = Metrics.recall_m(y_true, y_pred)
        return 2 * ((precision * recall) / (precision + recall + K.epsilon()))

    @staticmethod
    def dice_coef(y_true, y_pred, smooth=1):
        """
        Calculate the Dice coefficient metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.
            smooth: Smoothing parameter.

        Returns:
            Dice coefficient metric value.
        """
        y_true_f = K.flatten(y_true)
        y_pred_f = K.flatten(y_pred)
        intersection = K.sum(y_true_f * y_pred_f)
        return (2. * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)

    @staticmethod
    def dice_loss(y_true, y_pred, smooth=1):
        """
        Calculate the Dice loss metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.
            smooth: Smoothing parameter.

        Returns:
            Dice loss metric value.
        """
        intersection = K.sum(K.abs(y_true * y_pred), axis=-1)
        true_sum = K.sum(K.square(y_true), -1)
        pred_sum = K.sum(K.square(y_pred), -1)
        return 1 - ((2. * intersection + smooth) / (true_sum + pred_sum + smooth))

    @staticmethod
    def bce_dice_loss(y_true, y_pred):
        """
        Calculate the BCE-Dice loss metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.

        Returns:
            BCE-Dice loss metric value.
        """
        return Metrics.bce_loss(y_true, y_pred) + Metrics.dice_loss(y_true, y_pred)

    @staticmethod
    def bce_loss(y_true, y_pred):
        """
        Calculate the Binary Cross-Entropy (BCE) loss metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.

        Returns:
            BCE loss metric value.
        """
        return keras.losses.binary_crossentropy(y_true, y_pred, label_smoothing=0.2)

    @staticmethod
    def tversky(y_true, y_pred, smooth=1, alpha=0.7):
        """
        Calculate the Tversky index metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.
            smooth: Smoothing parameter.
            alpha: Weighting factor.

        Returns:
            Tversky index metric value.
        """
        y_true_pos = K.flatten(y_true)
        y_pred_pos = K.flatten(y_pred)
        true_pos = K.sum(y_true_pos * y_pred_pos)
        false_neg = K.sum(y_true_pos * (1 - y_pred_pos))
        false_pos = K.sum((1 - y_true_pos) * y_pred_pos)
        return (true_pos + smooth) / (true_pos + alpha * false_neg + (1 - alpha) * false_pos + smooth)

    @staticmethod
    def tversky_loss(y_true, y_pred):
        """
        Calculate the Tversky loss metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.

        Returns:
            Tversky loss metric value.
        """
        return 1 - Metrics.tversky(y_true, y_pred)

    @staticmethod
    def focal_tversky_loss(y_true, y_pred, gamma=0.75):
        """
        Calculate the focal Tversky loss metric.

        Args:
            y_true: Ground truth labels.
            y_pred: Predicted labels.
            gamma: Focusing parameter.

        Returns:
            Focal Tversky loss metric value.
        """
        return K.pow((1 - Metrics.tversky(y_true, y_pred)), gamma)

    @staticmethod
    def true_positives():
        """
        Create a metric for counting true positives.

        Returns:
            True positives metric.
        """
        return keras.metrics.TruePositives(name="tp")

    @staticmethod
    def false_positives():
        """
        Create a metric for counting false positives.

        Returns:
            False positives metric.
        """
        return keras.metrics.FalsePositives(name="fp")

    @staticmethod
    def true_negatives():
        """
        Create a metric for counting true negatives.

        Returns:
            True negatives metric.
        """
        return keras.metrics.TrueNegatives(name="tn")

    @staticmethod
    def false_negatives():
        """
        Create a metric for counting false negatives.

        Returns:
            False negatives metric.
        """
        return keras.metrics.FalseNegatives(name="fn")

    @staticmethod
    def binary_accuracy():
        """
        Create a metric for calculating binary accuracy.

        Returns:
            Binary accuracy metric.
        """
        return keras.metrics.BinaryAccuracy(name="accuracy")

    # check difference between this and precision_m output
    @staticmethod
    def precision():
        """
        Create a metric for calculating precision.

        Returns:
            Precision metric.
        """
        return keras.metrics.Precision(name="precision")

    @staticmethod
    def recall():
        """
        Create a metric for calculating recall.

        Returns:
            Recall metric.
        """
        return keras.metrics.Recall(name="recall")

    @staticmethod
    def auc():
        """
        Create a metric for calculating Area Under the Curve (AUC).

        Returns:
            AUC metric.
        """
        return keras.metrics.AUC(name="auc")

    @staticmethod
    def prc():
        """
        Create a metric for calculating Precision-Recall Curve (PRC).

        Returns:
            PRC metric.
        """
        return keras.metrics.AUC(name="prc", curve="PR")

    @staticmethod
    def one_hot_io_u(num_classes, name="one_hot_io_u"):
        """
        Create a metric for calculating Intersection over Union (IoU) using one-hot encoding.

        Args:
            num_classes: Number of classes.
            name: Name of the metric.

        Returns:
            One-hot IoU metric.
        """
        return keras.metrics.OneHotIoU(num_classes=num_classes, target_class_ids=list(range(num_classes)), name=name)



# -*- coding: utf-8 -*-

try:
    from aces.ee_utils import EEUtils
except ModuleNotFoundError:
    print("ModuleNotFoundError: Attempting to import from parent directory.")
    import os, sys
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")))

    from aces.ee_utils import EEUtils

import ee

EEUtils.initialize_session(use_highvolume=True)

country = "Bhutan"
region = "Punakha"
year = 2021

dem = ee.Image("MERIT/DEM/v1_0_3")

# baseModule = require("users/biplovbhandari/ACES_2:main.js")

planet_asia = ee.ImageCollection("projects/planet-nicfi/assets/basemaps/asia")

fc_country = ee.FeatureCollection(EEUtils.country_bbox(country))

admin_1 = ee.FeatureCollection("projects/servir-sco-assets/assets/Bhutan/BT_Admin_1")
paro = admin_1.filter(ee.Filter.eq("ADM1_EN", "Paro"))
punakha = admin_1.filter(ee.Filter.eq("ADM1_EN", "Punakha"))
punakha = punakha.map(lambda f: f.set("reduceId", 1))

monggar = admin_1.filter(ee.Filter.eq("ADM1_EN", "Monggar"))
monggar = monggar.map(lambda f: f.set("reduceId", 1))

name_region = {
    "Punakha": punakha,
    "Paro": paro,
    "Monggar": monggar
}

region_fc = name_region[region]

# Rice growing zone
rice_zone = dem.gt(1000).And(dem.lte(2600))
ic = planet_asia.filterBounds(region_fc.geometry()).filterDate(f"{year}-01-01", f"{year}-12-30")

# nlcms
# class number to name mapping
# 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
# water_body, snow, glacier, forest, riverbed, built_up,cropland, bare_soil, bare_rock, grassland, 
nlcms = ee.Image(f"projects/servir-hkh/RLCMS/HKH/landcover/hkh_landcover-{year}").clip(region_fc)

# 
buildings  = ee.FeatureCollection("projects/sat-io/open-datasets/MSBuildings/Bhutan")
buildings_layer = ee.Image().byte().paint(buildings, 1).unmask(0)

forest = nlcms.eq(4)
builtUp = nlcms.eq(6)

# Mosaic: Apr to Oct
img_mar = planet_asia.filterDate(ee.Date.fromYMD(year, 3, 1), ee.Date.fromYMD(year, 3, 1).advance(1, "month").advance(-1, "day")).first()
img_apr = planet_asia.filterDate(ee.Date.fromYMD(year, 4, 1), ee.Date.fromYMD(year, 4, 1).advance(1, "month").advance(-1, "day")).first()
img_may = planet_asia.filterDate(ee.Date.fromYMD(year, 5, 1), ee.Date.fromYMD(year, 5, 1).advance(1, "month").advance(-1, "day")).first()
img_jun = planet_asia.filterDate(ee.Date.fromYMD(year, 6, 1), ee.Date.fromYMD(year, 6, 1).advance(1, "month").advance(-1, "day")).first()
img_jul = planet_asia.filterDate(ee.Date.fromYMD(year, 7, 1), ee.Date.fromYMD(year, 7, 1).advance(1, "month").advance(-1, "day")).first()
img_aug = planet_asia.filterDate(ee.Date.fromYMD(year, 8, 1), ee.Date.fromYMD(year, 8, 1).advance(1, "month").advance(-1, "day")).first()
img_sep = planet_asia.filterDate(ee.Date.fromYMD(year, 9, 1), ee.Date.fromYMD(year, 9, 1).advance(1, "month").advance(-1, "day")).first()

rice_mosaic = ee.ImageCollection.fromImages(ee.List([img_apr, img_may, img_jun, img_jul, img_aug, img_sep])).median().mask(rice_zone)

none_masked_region = rice_mosaic.select(["B", "G"]).int32().reduceToVectors(
    reducer = ee.Reducer.firstNonNull(),
    geometry = region_fc,
    scale = 90,
    bestEffort = True,
    maxPixels = 1E13,
    tileScale = 16,
)

# Make the training dataset for k-means.
kMeans_training = rice_mosaic.sample(
    region = none_masked_region.geometry(),
    scale = 5,
    numPixels = 500,
    seed = 20,
    geometries = True,
)

# Instantiate the clusterer and train it.
# 7 has been an optimal class
num_k_class = 7
kMeans_clusterer = ee.Clusterer.wekaKMeans(num_k_class).train(kMeans_training)

# Cluster the input using the trained clusterer.
kMeans_cluster = ic.mosaic().cluster(kMeans_clusterer).updateMask(rice_zone)

# remap the clusters
from_class = [0, 1, 2, 3, 4, 5, 6]
# To class mapping
#  0 - cropland
#  1 - forest
#  2 - Built up
#  3 - water body, and others
# to_class = [0, 1, 2, 3, 0, 0, 1] # Paro # double check
to_class = [3, 1, 1, 0, 1, 1, 1] # Punakha

kMeans_cluster = kMeans_cluster.remap(from_class, to_class).rename("cluster")

kMeans_cluster = kMeans_cluster.where(nlcms.eq(1), ee.Image(3)) # class 1 // water
kMeans_cluster = kMeans_cluster.where(nlcms.eq(5), ee.Image(3)) # class 1 // river bed
kMeans_cluster = kMeans_cluster.where(nlcms.eq(4), ee.Image(1)) # class 4 // forest
kMeans_cluster = kMeans_cluster.where(nlcms.eq(6), ee.Image(2)) # class 6 // built-up

sample_kwargs = {
    "class_values": [0, 1, 2, 3],
    "class_points": [2000, 600, 600, 600],
}

stratified_points = EEUtils.generate_stratified_samples(
    kMeans_cluster, none_masked_region.geometry(),
    numPoints = 750,
    classBand = "cluster",
    **sample_kwargs
)

print("number of points per class", stratified_points.aggregate_histogram("cluster").getInfo())

export_kwargs = {
    "asset_id": "projects/servir-sco-assets/assets/Bhutan/ACES_2/stratifiedPoints_punakha_2021",
    "description": "stratifiedPoints_punakha_2021",
}

EEUtils.export_collection_data(stratified_points, export_type=["asset", "drive"], start_training=True, **export_kwargs)



# -*- coding: utf-8 -*-

import logging
logging.basicConfig(level=logging.INFO)
# logger = logging.getLogger(__name__)

import os
import datetime
import json
import pickle
from functools import partial

import numpy as np
import tensorflow as tf
from tensorflow import keras
from keras import callbacks

from aces.config import Config
from aces.metrics import Metrics
from aces.model_builder import ModelBuilder, DeSerializeInput, ReSerializeOutput, AddExtraFeatures
from aces.data_processor import DataProcessor
from aces.utils import Utils, TFUtils


class ModelTrainer:
    """
    A class for training, buidling, compiling, and running specified deep learning models.

    Attributes:
        config: An object containing the configuration settings for model training.
        model_builder: An instance of ModelBuilder for building the model.
        build_model: A partial function for building the model with the specified model type.
    """
    def __init__(self, config: Config, use_seed: bool = True, seed: int = 42):
        """
        Initialize the ModelTrainer object.

        Args:
            config: An object containing the configuration settings for model training.

        Attributes:
            config: The configuration settings for model training.
            model_builder: An instance of ModelBuilder for building the model.
            build_model: A partial function for building the model with the specified model type.
        """
        self.config = config
        self.use_seed = use_seed
        self.seed = seed
        # @FIXME: This isn't producing reproducable results
        if self.use_seed:
            # producable results
            import random
            logging.info(f"Using seed: {self.seed}")
            tf.random.set_seed(self.seed)
            np.random.seed(self.seed)
            random.seed(2)

        # @ToDO: Create a way to autoload loss function from the list without if else
        if self.config.LOSS == "custom_focal_tversky_loss":
            self.config.LOSS = Metrics.focal_tversky_loss
            self.LOSS_TXT = Metrics.focal_tversky_loss.__func__.__name__ # "focal_tversky_loss"
        else:
            self.config.LOSS_TXT = self.config.LOSS

        self.model_builder = ModelBuilder(
            features=self.config.FEATURES,
            out_classes=self.config.OUT_CLASS_NUM,
            optimizer=self.config.OPTIMIZER,
            loss=self.config.LOSS
        )
        self.build_model = partial(self.model_builder.build_model, model_type=self.config.MODEL_TYPE,
                                   **{"FOR_AI_PLATFORM": self.config.USE_AI_PLATFORM,
                                      "DERIVE_FEATURES": self.config.DERIVE_FEATURES})

    def train_model(self) -> None:
        """
        Train the model using the provided configuration settings.

        This method performs the following steps:

        1. Configures memory growth for TensorFlow.

        2. Creates TensorFlow datasets for training, testing, and validation.

        3. Builds and compiles the model.

        4. Prepares the output directory for saving models and results.

        5. Starts the training process.

        6. Evaluates and prints validation metrics.

        7. Saves training parameters, plots, and models.
        """
        logging.info("****************************************************************************")
        logging.info("****************************** Clear Session... ****************************")
        keras.backend.clear_session()
        logging.info("****************************************************************************")
        print(f"****************************** Configure memory growth... ************************")
        physical_devices = TFUtils.configure_memory_growth()
        self.config.physical_devices = physical_devices
        logging.info("****************************************************************************")
        logging.info("****************************** creating datasets... ************************")
        self.create_datasets(print_info=self.config.PRINT_INFO)

        if self.config.USE_AI_PLATFORM:
            logging.info("****************************************************************************")
            logging.info("******* building and compiling model for ai platform... ********************")
            self.build_and_compile_model_ai_platform()
        else:
            logging.info("****************************************************************************")
            logging.info("************************ building and compiling model... *******************")
            self.build_and_compile_model(print_model_summary=True)

        logging.info("****************************************************************************")
        logging.info("************************ preparing output directory... *********************")
        self.prepare_output_dir()
        logging.info("****************************************************************************")
        logging.info("****************************** training model... ***************************")
        self.start_training()

        if self.config.USE_AI_PLATFORM:
            logging.info(self.model.summary())

        logging.info("****************************************************************************")
        logging.info("****************************** evaluating model... *************************")
        self.evaluate_and_print_val()
        logging.info("****************************************************************************")
        logging.info("****************************** saving parameters... ************************")
        ModelTrainer.save_parameters(**self.config.__dict__)
        logging.info("****************************************************************************")
        logging.info("*************** saving model config and history object... ******************")
        self.save_history_object()
        if self.config.USE_AI_PLATFORM:
            ModelTrainer.save_model_config(self.config.MODEL_SAVE_DIR, **self._model.get_config())
        else:
            ModelTrainer.save_model_config(self.config.MODEL_SAVE_DIR, **self.model.get_config())
        logging.info("****************************************************************************")
        logging.info("****************************** saving plots... *****************************")
        self.save_plots()
        logging.info("****************************************************************************")
        logging.info("****************************** saving models... ****************************")
        self.save_models()
        logging.info("****************************************************************************")

    def prepare_output_dir(self) -> None:
        """
        Prepare the output directory for saving models and results.

        Creates a directory with a timestamped name and increments the version number if necessary.
        """
        if not self.config.AUTO_MODEL_DIR_NAME:
            self.config.MODEL_SAVE_DIR = self.config.OUTPUT_DIR / self.config.MODEL_DIR_NAME
            logging.info(f"> Saving models and results at {self.config.MODEL_SAVE_DIR}...")
            os.mkdir(self.config.MODEL_SAVE_DIR)
        else:
            today = datetime.date.today().strftime("%Y_%m_%d")
            iterator = 1
            while True:
                model_dir_name = f"trial_{self.config.MODEL_TYPE}_{today}_v{iterator}"
                self.config.MODEL_SAVE_DIR = self.config.OUTPUT_DIR / model_dir_name
                try:
                    os.mkdir(self.config.MODEL_SAVE_DIR)
                except FileExistsError:
                    logging.info(f"> {self.config.MODEL_SAVE_DIR} exists, creating another version...")
                    iterator += 1
                    continue
                else:
                    logging.info(f"> Saving models and results at {self.config.MODEL_SAVE_DIR}...")
                    break

    def create_datasets(self, print_info: bool = False) -> None:
        """
        Create TensorFlow datasets for training, testing, and validation.

        Args:
            print_info: Flag indicating whether to print dataset information.

        Prints information about the created datasets if print_info is set to True.
        """
        self.TRAINING_DATASET = DataProcessor.get_dataset(
            f"{str(self.config.TRAINING_DIR)}/*",
            self.config.FEATURES,
            self.config.LABELS,
            self.config.PATCH_SHAPE[0],
            self.config.BATCH_SIZE,
            self.config.OUT_CLASS_NUM,
            **{**self.config.__dict__, "training": True},
        ).repeat()

        self.VALIDATION_DATASET = DataProcessor.get_dataset(
            f"{str(self.config.VALIDATION_DIR)}/*",
            self.config.FEATURES,
            self.config.LABELS,
            self.config.PATCH_SHAPE[0],
            1,
            self.config.OUT_CLASS_NUM,
            **self.config.__dict__,
        ).repeat()

        self.TESTING_DATASET = DataProcessor.get_dataset(
            f"{str(self.config.TESTING_DIR)}/*",
            self.config.FEATURES,
            self.config.LABELS,
            self.config.PATCH_SHAPE[0],
            1,
            self.config.OUT_CLASS_NUM,
            **self.config.__dict__,
        )

        if print_info:
            logging.info("Printing dataset info:")
            DataProcessor.print_dataset_info(self.TRAINING_DATASET, "Training")
            DataProcessor.print_dataset_info(self.TESTING_DATASET, "Testing")
            DataProcessor.print_dataset_info(self.VALIDATION_DATASET, "Validation")

    def build_and_compile_model(self, print_model_summary: bool = True) -> None:
        """
        Build and compile the model.

        Args:
            print_model_summary: Flag indicating whether to print the model summary.

        Builds and compiles the model using the provided configuration settings.

        Prints the model summary if print_model_summary is set to True.
        """
        self.model = self.build_model(**self.config.__dict__)
        if print_model_summary:  logging.info(self.model.summary())

    def build_and_compile_model_ai_platform(self) -> None:
        """
        Build and compile the model.

        Args:
            print_model_summary: Flag indicating whether to print the model summary.

        Builds and compiles the model using the provided configuration settings.

        Prints the model summary if print_model_summary is set to True.
        """
        model, wrapped_model = self.build_model(**self.config.__dict__)
        logging.info(model.summary())
        self._model = model
        self.model = wrapped_model

    def start_training(self) -> None:
        """
        Start the training process.

        Trains the model using the provided configuration settings and callbacks.
        """
        model_checkpoint = callbacks.ModelCheckpoint(
            f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_CHECKPOINT_NAME}",
            monitor=self.config.CALLBACK_PARAMETER,
            save_best_only=True,
            mode="auto",
            verbose=1,
            save_weights_only=False,
        )  # save best model

        early_stopping = callbacks.EarlyStopping(
            monitor=self.config.CALLBACK_PARAMETER,
            patience=int(0.3 * self.config.EPOCHS),
            verbose=1,
            mode="auto",
            restore_best_weights=True,
        )

        tensorboard = callbacks.TensorBoard(log_dir=str(self.config.MODEL_SAVE_DIR / "logs"), write_images=True)

        def lr_scheduler(epoch):
            if epoch < self.config.RAMPUP_EPOCHS:
                return self.config.MAX_LR
            elif epoch < self.config.RAMPUP_EPOCHS + self.config.SUSTAIN_EPOCHS:
                return self.config.MID_LR
            else:
                return self.config.MIN_LR

        lr_callback = callbacks.LearningRateScheduler(lambda epoch: lr_scheduler(epoch), verbose=True)

        model_callbacks = [model_checkpoint, tensorboard]

        if self.config.USE_ADJUSTED_LR:
            model_callbacks.append(lr_callback)

        if self.config.EARLY_STOPPING:
            model_callbacks.append(early_stopping)

        self.model_callbacks = model_callbacks

        self.history = self.model.fit(
            x=self.TRAINING_DATASET,
            epochs=self.config.EPOCHS,
            steps_per_epoch=(self.config.TRAIN_SIZE // self.config.BATCH_SIZE),
            validation_data=self.VALIDATION_DATASET,
            validation_steps=self.config.VAL_SIZE,
            callbacks=model_callbacks,
        )

        # either save the wrapped model or the original model
        # named as "trained-model" to avoid confusion
        # self.model.save(f"{self.config.MODEL_SAVE_DIR}/trained-wrapped-model")
        self.model.save(f"{self.config.MODEL_SAVE_DIR}/trained-model")

    def evaluate_and_print_val(self) -> None:
        """
        Evaluate and print validation metrics.

        Evaluates the model on the validation dataset and prints the metrics.
        """
        logging.info("************************************************")
        logging.info("************************************************")
        logging.info("Validation")
        # Tip: You can remove steps=self.config.TEST_SIZE and match the TEST_SIZE from the env
        evaluate_results = self.model.evaluate(self.TESTING_DATASET) # , steps=self.config.TEST_SIZE
        with open(f"{self.config.MODEL_SAVE_DIR}/evaluation.txt", "w") as evaluate:
            evaluate.write(json.dumps(dict(zip(self.model.metrics_names, evaluate_results))))
        for name, value in zip(self.model.metrics_names, evaluate_results):
            logging.info(f"{name}: {value}")
        logging.info("\n")

    @staticmethod
    def save_parameters(**config) -> None:
        """
        Save the training parameters to a text file.

        Saves the training parameters used in the configuration settings to a text file.
        """
        with open(f"{str(config.get('MODEL_SAVE_DIR'))}/parameters.txt", "w") as f:
            f.write(f"TRAIN_SIZE: {config.get('TRAIN_SIZE')}\n")
            f.write(f"TEST_SIZE: {config.get('TEST_SIZE')}\n")
            f.write(f"VAL_SIZE: {config.get('VAL_SIZE')}\n")
            f.write(f"BATCH_SIZE: {config.get('BATCH_SIZE')}\n")
            f.write(f"EPOCHS: {config.get('EPOCHS')}\n")
            f.write(f"LOSS: {config.get('LOSS_TXT')}\n")
            f.write(f"TRAINING_DIR: {config.get('TRAINING_DIR')}\n")
            f.write(f"TESTING_DIR: {config.get('TESTING_DIR')}\n")
            f.write(f"VALIDATION_DIR: {config.get('VALIDATION_DIR')}\n")
            if config.get('USE_ADJUSTED_LR'):
                f.write(f"USE_ADJUSTED_LR: {config.get('USE_ADJUSTED_LR')}\n")
                f.write(f"MAX_LR: {config.get('MAX_LR')}\n")
                f.write(f"MID_LR: {config.get('MID_LR')}\n")
                f.write(f"MIN_LR: {config.get('MIN_LR')}\n")
                f.write(f"RAMPUP_EPOCHS: {config.get('RAMPUP_EPOCHS')}\n")
                f.write(f"SUSTAIN_EPOCHS: {config.get('SUSTAIN_EPOCHS')}\n")
            f.write(f"DROPOUT_RATE: {config.get('DROPOUT_RATE')}\n")
            f.write(f"ACTIVATION_FN: {config.get('ACTIVATION_FN')}\n")
            f.write(f"FEATURES: {config.get('FEATURES')}\n")
            f.write(f"LABELS: {config.get('LABELS')}\n")
            f.write(f"PATCH_SHAPE: {config.get('PATCH_SHAPE')}\n")
            f.write(f"CALLBACK_PARAMETER: {config.get('CALLBACK_PARAMETER')}\n")
            f.write(f"MODEL_TYPE: {config.get('MODEL_TYPE')}\n")
            f.write(f"TRANSFORM_DATA: {config.get('TRANSFORM_DATA')}\n")
            f.write(f"MODEL_NAME: {config.get('MODEL_NAME')}.h5\n")
            f.write(f"MODEL_CHECKPOINT_NAME: {config.get('MODEL_CHECKPOINT_NAME')}.h5\n")
        f.close()

    @staticmethod
    def save_model_config(save_dir, **model_config) -> None:
        with open(f"{save_dir}/config.json", "w") as f:
            json.dump(model_config, f, indent=4)
        f.close()

    def save_plots(self) -> None:
        """
        Save plots and model visualization.

        Saves the model architecture plot, training history plot, and model object.
        """
        logging.info(f"Saving plots and model visualization at {self.config.MODEL_SAVE_DIR}...")
        if self.config.USE_AI_PLATFORM:
            keras.utils.plot_model(self._model, f"{self.config.MODEL_SAVE_DIR}/model.png", show_shapes=True, rankdir="TB")
            keras.utils.plot_model(self.model, f"{self.config.MODEL_SAVE_DIR}/wrapped_model.png", show_shapes=True, rankdir="LR") # rankdir='TB'
        else:
            keras.utils.plot_model(self.model, f"{self.config.MODEL_SAVE_DIR}/model.png", show_shapes=True, rankdir="TB") # rankdir='TB'
        Utils.plot_metrics([key.replace("val_", "") for key in self.history.history.keys() if key.startswith("val_")],
                           self.history.history, len(self.history.epoch), self.config.MODEL_SAVE_DIR)

    def save_history_object(self) -> None:
        """
        Save the history object.
        """
        with open(f"{self.config.MODEL_SAVE_DIR}/model.pkl", "wb") as f:
            pickle.dump(self.history.history, f)

        with open(f"{self.config.MODEL_SAVE_DIR}/model.txt", "w") as f:
            f.write(json.dumps(self.history.history))


    def load_and_save_models(self) -> None:
        """
        Load the trained models.

        Loads the trained models from different formats: h5 and tf formats.
        """
        self.config.MODEL_SAVE_DIR = self.config.OUTPUT_DIR / self.config.MODEL_DIR_NAME
        self.model = tf.keras.models.load_model(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_CHECKPOINT_NAME}.tf")
        updated_model = self.serialize_model()
        # if not issubclass(self.model.__class__, keras.Model):
        #     # Saving the model to HDF5 format requires the model to be a Functional model or a Sequential model
        #     updated_model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}.h5", save_format="h5")
        updated_model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}.h5", save_format="h5")
        updated_model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}.tf", save_format="tf")


    def save_models(self) -> None:
        """
        Save the trained models.

        Saves the trained models in different formats: h5 and tf formats.
        """
        if self.config.USE_AI_PLATFORM:
            updated_model = self.serialize_model()
            # updated_model = self.model

            # if not issubclass(self.model.__class__, keras.Model):
            #     # Saving the model to HDF5 format requires the model to be a Functional model or a Sequential model
            #     updated_model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}.h5", save_format="h5")

            # updated_model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}.h5", save_format="h5")
            # updated_model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}.tf", save_format="tf")
            updated_model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}")
        else:
            if not issubclass(self.model.__class__, keras.Model):
                self.model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}.h5", save_format="h5")
            self.model.save(f"{str(self.config.MODEL_SAVE_DIR)}/{self.config.MODEL_NAME}", save_format="tf")

    def serialize_model(self) -> tf.keras.Model:
        """
        Serialize and save the trained models.

        Saves the trained models in different formats: h5 and tf formats.
        """
        input_deserializer = DeSerializeInput(self.config.FEATURES)
        output_deserializer = ReSerializeOutput()
        serialized_inputs = {
            b: tf.keras.Input(shape=[], dtype="string", name=b) for b in self.config.FEATURES
        }
        updated_model_input = input_deserializer(serialized_inputs)
        updated_model = self.model(updated_model_input)
        updated_model = output_deserializer(updated_model, "output")
        updated_model = tf.keras.Model(serialized_inputs, updated_model)
        keras.utils.plot_model(updated_model, f"{self.config.MODEL_SAVE_DIR}/serialized_model.png", show_shapes=True, rankdir="LR")
        return updated_model
